/*
 * cancan  --  mud client with telnet protocol
 *
 * written by Mattias Engdegrd (Yorick) (f91-men@nada.kth.se) 1992-95
 * Maintainers: Ilie (finnag@pvv.unit.no) (and Yorick)
 * originally inspired by the Tintin client by Peter Unold but contains no
 * Tintin code.
 * Also greatly improved upon by Vivriel and Thuzzle. As usual, all the
 * developers are in debt to countless users for suggestions and debugging.
 *
 * This program is in the public domain.
 */

#define VERSION "2.6.7 (Yorick, Vivriel, Thuzzle, Ilie 1992-99)"
#define SAVEFILEVER "2.6"

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/time.h>
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <signal.h>

int select();
int gettimeofday();

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "ctypes.h"
#include "cmd.h"
#include "beam.h"

/*	function declarations		*/
static void init_def_file();
static void mainlupe();
static void process_attn();
#ifndef OMIT_SCHEDULE
static struct timeval *check_schedule();
#endif
#ifdef MOTDFILE
static void printmotd();
#endif
static void get_user_input();
static void get_remote_output();
#ifndef OMIT_SESSIONS
static void process_sub_input();
#endif
static void process_input();
static void setsigs();
static void read_settings();
static void write_settings();
void termsig();
void pipesig();
void childsig();
void winchsig();

/*	GLOBALS		*/

char *hostname;
char *portnumber;
char *initstr;		/* initial command to run (send) on connect */

FILE *capfile;		/* capture file or NULL */

#ifndef OMIT_SESSIONS
/* session data indexed by socket descriptor */
static char *sess_id[MAXFDSCAN];	/* session name */
static int sess_active[MAXFDSCAN];	/* active flag */
int main_sockfd;	/* main session socket fd */
#endif

int sockfd;		/* current socket file descriptor */

int attention;		/* special attention flags (ATTN_* in cancan.h) */
int linemode;		/* current line mode (LM_* in cancan.h) */

fd_set fdset;		/* set of descriptors to select() on */

/* booleans */
char current_options[] = OPTION_DEFAULTS;
int mp_active;		/* 1 = we have #identified ourselves */
int expect_prompt;	/* are we waiting for remote data? */
int new_settings;	/* set this when user defines new alias/action/etc */
int verbose;		/* this affects verbosity of #commands */

int flashback;		/* cursor is on excursion, put back there */
char *footnote;		/* displayed at the bottom (just once) */
char *prefixstr;	/* inserted in the editing buffer each time */
char insertbuf[LINELEN];/* used in many ways. hm */
char promptstr[LINELEN];/* entire prompt string (for redrawing) */


int main(argc, argv)
int argc; char *argv[];
{
    static char outbuf[BUFSIZE];
    if(argc < 2 || argc > 3) {
	fprintf(stderr, "usage:\t%s host port\nor\t%s definition-file\n",
		argv[0], argv[0]);
	return 1;
    }
    /* full buffered stdout means less write() calls */
    setvbuf(stdout, outbuf, _IOFBF, sizeof(outbuf));
    init_env();		/* set up our environment first */
    init_ctype();
#ifndef VT100
    init_termcap();
#endif
#ifdef TELNETBUG
    set_stdcolour(DEFAULT_ATTRCODE);
#endif
#ifndef OMIT_PRINTVER
    printver();
#endif
#ifdef MOTDFILE
    printmotd();
#endif
    init_commands();
    init_completions();
    default_bindings();
    if(argc == 2) {	/* one arg, assuming it is definition file name */
	init_def_file(argv[1]);
    } else {	/* two args: assume they are hostname and port number */
	hostname = argv[1];	/* should be safe not to str_dup these */
	portnumber = argv[2];
	keypad_bindings();
    }
    set_terminal();
    setsigs();
    mainlupe();
    return 1;
}

/*
 * initialize cancan from definition file or create one
 */
static void init_def_file(df)
char *df;
{
    char buf[LINELEN], *p;
    FILE *f;
    /*
     * rules: if cancan dir is set it is searched first.
     *        if file doesn't exist, it is created there.
     * If a slash appears in the name, the cancan dir isn't used.
     */
    if(!strchr(df, '/') && (p = get_env("CANCANDIR")) && *p) {
	/* BUG: buffer overflow possible. <snicker> */
	sprintf(buf, "%s/%s", p, df);
	if((f = fopen(buf, "r")) || !(f = fopen(df, "r")))
	    df = buf;
    } else
	f = fopen(df, "r");
    put_env("SAVEFILE", df);
    if(f) {
	read_settings(f, buf);
	if(ferror(f)) {
	    syswarn("Can't read definition file");
	    exit(1);
	}
	fclose(f);
    } else {
	if(!(f = fopen(df, "a+"))) {
	    syswarn("Cannot open definition file");
	    exit(1);
	}
	printf("Creating %s\nHost name  :", df);
	fflush(stdout); fgets(buf, sizeof(buf), stdin);
	if(*buf != '\n') {
	    chop_str(buf, '\n'); hostname = str_dup(buf);
	    printf("Port number:");
	    fflush(stdout); fgets(buf, sizeof(buf), stdin);
	    chop_str(buf, '\n'); portnumber = str_dup(buf);
	}
	keypad_bindings();
	write_settings(f, buf);
	fclose(f);
    }
}

/*
 * save cancan settings. mode == 1 for manual and 0 for autosave.
 */
void save_def_file(mode)
int mode;
{
    char buf[LINELEN], *df = get_env("SAVEFILE");
    FILE *f;
    /*
     * NOTE: backup file is an extra copy, not a previous version of
     * the savefile. (I wish there was a syscall to swap around two
     * dirents' inode numbers.)
     */
    if(!df) {
	if(mode)
	    printf("#not saved: please set the SAVEFILE variable\n");
	return;
    } else {
	if((f = fopen(df, "w"))) {
	    write_settings(f, buf);
	    if(ferror(f) | fclose(f))
		{ syswarn("#unable to save definitions"); return; }
	} else
	    { syswarn("#can't write to definition file"); return; }
    }
    if(mode && OPTION(OPT_BACKUPFILE)) {
	sprintf(buf, "%s~", df);
	if((f = fopen(buf, "w"))) {
	    write_settings(f, buf);
	    if(ferror(f) | fclose(f))
		{ syswarn("#error writing backup file"); }
	} else
	    { syswarn("#cannot write to backup file"); }
    }
}

/*
 * show current version
 */
void printver()
{
    printf(
#ifdef __STDC__
	"cancan ver " VERSION "\nCompiled " __DATE__ " (%s %s)\n",
#else
	"cancan ver %s (%s %s)\n", VERSION,
#endif
#ifdef VT100
	"vt100-only",
#else
	"termcaps",
#endif
#if !defined(USE_TERMIO) && !defined(USE_SGTTY)
	"termios"
#endif
#ifdef USE_TERMIO
	"termio"
#endif
#ifdef USE_SGTTY
	"sgtty"
#endif
    );
}

#ifdef MOTDFILE
/*
 * print the message of the day if present
 */
static void printmotd()
{
    char line[LINELEN];
    FILE *f = fopen(MOTDFILE, "r");
    if(f) {
	while(fgets(line, sizeof(line), f))
	    fputs(line, stdout);
	fclose(f);
    }
}
#endif

/*
 * main loop.
 */
static void mainlupe()
{
    int err;
    fd_set readfds;
    struct timeval *timeout;
    static struct timeval timeoutbuf;

    FD_ZERO(&fdset);
    FD_SET(0, &fdset);		/* standard input (terminal) */
#ifndef OMIT_SESSIONS
    if(hostname && connect_sess("main", (char *)0, (char *)0) < 0)
	exitcancan("");
#else
    if(!hostname || (sockfd = do_connect(hostname, portnumber)) < 0)
	exitcancan("#please specify correct hostname and port number\n");
    FD_SET(sockfd, &fdset);	/* add the socket to select() set */
#endif
    if(initstr) parse(initstr, 0, (struct param *)0);

    printf("Type #help for help.\n");

    /* make sure the input line is always the bottom one */
    check_winsize();
    lastline();

    /*		main lupe	*/

    for(;;) {

#ifndef OMIT_SESSIONS
	sockfd = main_sockfd;
#endif

#ifndef OMIT_SCHEDULE
	timeout = check_schedule(&timeoutbuf);
#else
	timeout = 0;
#endif

	if(attention) {
	    if(attention & ATTN_FLASHBACK)
#ifndef OMIT_SCHEDULE
		if(!timeout || timeout->tv_sec)
#endif
		{
		    timeout = &timeoutbuf;
		    timeout->tv_sec = FLASHDELAY / 1000000;
		    timeout->tv_usec = FLASHDELAY % 1000000;
		}
	    if(attention & ATTN_FOOTNOTE)
		display_footnote();
	}

	flush_output_buffer();	/* sploooosh */
	fflush(stdout);		/* ditto */

	/*
	 * NOTE: we should pass highest_fd + 1, rather than MAXFDSCAN
	 * NOTE: check_schedule gettimeofday should come right before select
	 * NOTE: a signal arriving just before we enter select has to await
	 * some other event, before it gets processed... is there any way
	 * around this without chemistry?
	 */
	COPY_FDS(&fdset, &readfds);
	err = select(MAXFDSCAN, &readfds, (fd_set *)0, (fd_set *)0, timeout);
	if(err < 0 && errno != EINTR) syserr("select");

	if(attention) process_attn();

	if(err > 0) {
#ifndef OMIT_SESSIONS
	    if(FD_ISSET(0, &readfds)) {
		get_user_input();
		if(!--err) continue;
	    }
	    do if(FD_ISSET(sockfd, &readfds)) {
		get_remote_output();
		if(!--err) break;
	    } while(++sockfd < MAXFDSCAN);
#else
	    if(FD_ISSET(0, &readfds))
		get_user_input();
	    if(FD_ISSET(sockfd, &readfds))
		get_remote_output();
#endif
	}
    }
}

static void process_attn()
{
    int attn = attention;
    attention ^= attn;
    if(attn & ATTN_FLASHBACK) {
	moveto(flashback);
    }
    if(attn & ATTN_FOOTNOTE) {
	str_assign(&footnote, (char *)0);
	display_footnote();
    }
    if(attn & ATTN_WINCHSIG) {
	signal(SIGWINCH, winchsig);
	if(check_winsize()) {
	    clear_all();
	    redraw_all();
	}
    }
    if(attn & ATTN_CHILDSIG) {
	/*
	 * NOTE: on some systems (e.g. AIX), setting a sigchld handler
	 * will promptly deliver the signal if waitpid() has anything to
	 * report. Thus, we would get called once more on the next loop,
	 * and do extra signal() and waitpid() calls. Better flush one
	 * child before reinstalling the signal handler.
	 */
	attend_child(0);
	signal(SIGCHLD, childsig);
	attend_child(-1);
    }
}

#ifndef OMIT_SCHEDULE
/*
 * process and update delayed commands list.
 */
static struct timeval *check_schedule(to)
struct timeval *to;
{
    static struct timeval time0;
    if(schedule) {
	long delta = -time0.tv_sec;
	/*
	 * NOTE: gettimeofday is somewhat expensive. However, even if we
	 * used alarm timer, the schedule still needs to be up to date,
	 * in case user decides to list it.
	 */
	gettimeofday(&time0, (struct timezone *)0);
	time0.tv_sec = SCH_HZ * time0.tv_sec +
		time0.tv_usec / (1000000 / SCH_HZ);
	if(delta) delta += time0.tv_sec;
	/*
	 * NOTE: when a command is run while charmode is active, the
	 * output produced (if any) might not look too good since
	 * LF to CRLF mapping is not in effect.
	 * TODO: let the user freeze the schedule, do smart recovery
	 * when command is long overdue.
	 */
	do {
	    if(schedule->delta > delta) {
		delta = (schedule->delta -= delta);
		to->tv_sec = delta / SCH_HZ;
		to->tv_usec = (1000000 / SCH_HZ) * (delta % SCH_HZ);
		return to;
	    } else if(schedule->delta >= 0) {
		delta -= schedule->delta;
		schedule->delta = -1;
		if(schedule->next)
		    schedule->next->delta++;
		run_command(schedule->commands);
	    } else
		delete_schedulenode(&schedule);
	} while(schedule);
    }
    time0.tv_sec = 0;
    return 0;
}
#endif

/*
 * read a single character from stdin.
 * timeout == 0 means wait indefinitely (no timeout).
 * return 1 if a char was read, or 0 if not.
 */
int get_one_char(cp, timeout)
char *cp;
int timeout;
{
    int err;
    fd_set fds;
    struct timeval timeoutbuf, *to;
    to = timeout ? &timeoutbuf : 0;
    do {
	FD_ZERO(&fds);	/* this is probably redundant */
	FD_SET(0, &fds);
	timeoutbuf.tv_sec = KBD_TIMEOUT / 1000000;
	timeoutbuf.tv_usec = KBD_TIMEOUT % 1000000;
	err = select(1, &fds, (fd_set *)0, (fd_set *)0, to);
    } while(err < 0 && errno == EINTR);
    if(err < 0) syserr("select");
    if(err && read(0, cp, 1) != 1)
	syserr("read from stdin");
    return err;
}

/*
 * read terminal input and handle it.
 */
static void get_user_input()
{
    char rbuf[CAPLEN];
    int n = read(0, rbuf, sizeof(rbuf));
    if(n > 0)
	line_edit(rbuf, n);
    else
	syserr("read from stdin");
}

/*
 * get output from the socket and process/display it.
 */
static void get_remote_output()
{
    /*
     * NOTE: a typical screenful of text might be anywhere from
     * two to four k in size. We should buffer at least this much.
     */
    static char buffer[LINELEN];
    int n = read_socket(buffer, (int)sizeof(buffer) - 1);
    /* leave room for terminating zero:          ^^^ */
    if(n <= 0) {
#ifndef OMIT_SESSIONS
	/* NOTE: What if we are in charmode??  */
	clear_all();
	disconnect_sess(sockfd);
	redraw_all();
	return;
#else
	exitcancan("#connection closed by remote host.\n");
#endif
    }

#ifndef OMIT_SESSIONS
    if(!sess_active[sockfd]) {
	n = tn_process(buffer, n);
	return;
    } else if(!is_main_sess()) {
	clear_all();
	n = tn_process(buffer, n);
	process_sub_input(buffer, n);
	redraw_all();
    } else
#endif
    if(linemode != LM_CHAR) {
	clear_input();
	n = tn_process(buffer, n);
	process_input(buffer, n);
	redraw_input();
    } else {
	n = tn_process(buffer, n);
	to_capfile(buffer, n, 0);
	fwrite(buffer, 1, n, stdout);
    }
}

/*
 * process input from remote host:
 * detect special sequences, trigger actions, locate prompt, word-wrap,
 * print to stdout
 * TODO: use a linebuffer (== prompt). tn_getline(buf, n, lbuf) ?
 * assert(sizeof(buf) >= size + 1)
 */
static void process_input(buf, size)
char *buf; int size;
{
    static editsess *esess;
    char *p, *tail;
    int n, i = 0;

    if(expect_prompt) {
	/* NOTE: expect_prompt is zeroed even if we get only a mp message */
	expect_prompt = 0;
	if(esess) esess->mode = 0;	/* sending cancels the session */
    }
    while(i < size) {
	p = buf + i;
	n = size - i;
	if(esess || (p[0] == MPI[0] && mp_active && !col0 && IS_MPI(p, n)))
	    i += mp_receive(&esess, p, n);
	else {
	    /* process block one line at a time: */
	    if(*p == '\r' && !col0)			/* leading CR */
		{ if(++i >= size) break; }
	    for(n = i; i < size; i++)			/* scan for LF */
		if(buf[i] == '\n') break;
	    p = buf + n; n = i - n;
	    if(n && p[n - 1] == '\r') n--;		/* trailing CR */
	    p[n] = 0;					/* zero terminate */
	    to_capfile(p, n, (i < size));
	    if(i < size) {
		i++;
		if(!search_action(p, n))
		    continue;				/* line #gagged */
		if(n) wrap_out(p, n);
		newline(); col0 = 0;
	    } else {
		/*
		 * FEATURE: action search on line fragments can give results
		 * differing from a search on a complete line. Alternatively,
		 * the search could be postponed until a newline is read.
		 */
		search_action(p, n);
		if(!col0) promptstr[0] = 0;
		tail = wrap_out(p, n);
		if(tail)
		    { p = tail; promptstr[0] = 0; }
		if(col0 < cols_1) {
		    n = str_len(promptstr);
		    if(n + str_len(p) < LINELEN) {
			strcpy(promptstr + n, p);	/* promptocatenate */
			break;
		    }
		}
		newline(); col0 = 0;
		break;
	    }
	}
    } /* end while */
}

#ifndef OMIT_SESSIONS
/*
 * process input from remote host (subsidiary session):
 * FEATURE: no wrapping, no marking, or triggers in subsession
 * assert(sizeof(buf) >= size + 1)
 */
static void process_sub_input(buf, size)
char *buf; int size;
{
    int i;
    for(i = 0; i < size; i++) {
	char *id, *line, *s = buf + i;
	if(*s == '\r') continue;
	for(line = s; i < size && *s != '\n'; s++, i++)
	    if(!isprint(*s)) *s = ' ';	/* weed out unprintables */
	*s = 0;
	id = sess_id[sockfd];
	printf("##%s> %.*s", id, cols_1 - (4 + str_len(id)), line);
	newline();
    }
}
#endif

#ifndef NOSHELL
/*
 * execute a shell command and return the result
 */
int execute(cmd)
char *cmd;
{
    int pid = do_fork();
    if(!pid) {
	static char *args[] = { "/bin/sh", "-c", 0, 0 };
	args[2] = *cmd ? cmd : args[0];		/* sh exec sh? hm */
	do_execve(args);
    }
    if(pid > 0)
	pid = attend_child(pid);
    return pid;		/* return success/error status */
}
#endif

/*
 * set up our signal handlers
 */
static void setsigs()
{
    signal(SIGTERM, termsig);
    signal(SIGPIPE, pipesig);
    signal(SIGCHLD, childsig);
    signal(SIGWINCH, winchsig);
}

void winchsig()
{
    attention |= ATTN_WINCHSIG;
}

void childsig()
{
    attention |= ATTN_CHILDSIG;
}

void pipesig()
{
    exitcancan("\n#broken pipe.\n");
}

void termsig()
{
    exitcancan("\n#termination signal.\n");
}

void intrsig()
{
    if(sockfd) do_disconnect(sockfd);
    exitcancan("\n#interrupted.\n");
}

/*
 * suspend ourselves
 */
void suspend_cancan()
{
    if(signal(SIGTSTP, SIG_IGN) == SIG_DFL) {
	skip_edline();
	reset_terminal();
	fflush(stdout);
	signal(SIGTERM, SIG_IGN);
	signal(SIGTSTP, SIG_DFL);
	kill(getpid(), SIGTSTP);
	set_terminal();
	signal(SIGTERM, termsig);
	check_winsize();	/* in case size changed meanwhile */
	lastline();		/* so we know where the cursor is */
	redraw_all();
    } else	/* NOTE: could spawn a subshell if cannot suspend. */
	to_footnote("#your shell probably lacks job control.");
}

/*
 * terminate cancan as cleanly as possible
 */
void exitcancan(message)
char *message;
{
    lastline();
    reset_terminal();
    printf("%s#cancan stopped dancing.\n", message);
    exit(0);
}

/*
 * print an error message
 */
void syswarn(msg)
char *msg;
{
    int err = errno;
#ifdef USE_STRERROR
    printf("%s (%s)\n", msg, strerror(err));
#else
    if(err > 0 && err < sys_nerr)
	printf("%s (%s)\n", msg, sys_errlist[err]);
    else
	printf("%s\n", msg);
#endif
}

/*
 * print system call error message and terminate
 */
void syserr(msg)
char *msg;
{
    int err = errno;
    reset_terminal();
#ifdef USE_STRERROR
    fprintf(stderr, "\n#fatal system call error: %s (%s)\n",
	    msg, strerror(err));
#else
    fprintf(stderr, "\n#fatal system call error: %s (", msg);
    if(err > 0 && err < sys_nerr)
	fprintf(stderr, "%s)\n", sys_errlist[err]);
    else
	fprintf(stderr, "%d)\n", err);
#endif
    exit(1);
}

void to_footnote(msg)
char *msg;
{
    attention |= ATTN_FOOTNOTE;
    str_assign(&footnote, msg);
}

void to_capfile(str, len, nl)
char *str; int len;
int nl;
{
    if(capfile) {
	fwrite(str, 1, len, capfile);
	if(nl) putc('\n', capfile);
    }
}

/*
 * read definitions from file
 */
static void read_settings(f, buf)
FILE *f; char *buf;
{
    typedef enum {
	def_action, def_alias, def_bind, def_mark, def_hilite,
	def_option, def_init, def_host, def_savefile, def_error
    } def_token;
    static char * const def_names[] = {
	"#action", "#alias", "#bind", "#mark", "#hilite",
	"#option", "#init", "#host", "#savefile", 0
    };
    static cmd_func * const def_parsers[] = {
	parse_action, parse_alias, parse_bind, parse_marker, parse_hilite
    };
    char *line, *args;
    int n_err = 0, ln = 0;
    def_token what;

    while(++ln, fgets(buf, LINELEN, f)) {
	if(*buf == '\n') continue;
	chop_str(line = buf, '\n');
	what = (def_token)get_token(&line, def_names);
	if(!(args = line))
	    what = def_error;

	if(what < def_option) {
	    new_settings = 0;		/* just a hack to catch errors */
	    (*def_parsers[what])(args);
	    if(new_settings) continue;
	    printf("Bad %s definition, line %d\n", def_names[what], ln);
	} else if(what == def_option) {
	    parse_options(args); continue;
	} else if(what == def_init) {
	    str_assign(&initstr, args); continue;
	} else if(what == def_savefile) {
	    if(!strcmp(args, SAVEFILEVER)) continue;
	    printf("Warning: expected definition file version is '%s'\n",
		    SAVEFILEVER);
	} else {
	    if(what == def_host) {
		char *rest = chop_str(args, ' ');
		if(rest) {
		    str_assign(&portnumber, rest);
		    str_assign(&hostname, args);
		    continue;
		}
	    }
	    printf("Bad definition file format, line %d\n", ln);
	}

	if(++n_err < 10) continue;
	printf("Too many errors encountered, quitting.\n");
	exit(1);
    }
    if(n_err) {
	/* printf("Errors were encountered; autosave turned off.\n"); */
	OPTION(OPT_AUTOSAVE) = 0;
    }
}

/*
 * store definitions to file
 */
static void write_settings(f, buf)
FILE *f; char *buf;
{
    aliasnode *al;
    actionnode *ac;
    marknode *m;
    keynode *k;

    fprintf(f, "#savefile %s\n", SAVEFILEVER);
    print_options(buf);
    fprintf(f, "%s\n", buf);
    if(edattr != NOATTRCODE)
	fprintf(f, "#hilite %s\n", attr_name(edattr));
    if(hostname)
	fprintf(f, "#host %s %s\n", hostname, portnumber);
    if(initstr)
	fprintf(f, "#init %s\n", initstr);
    for(al = aliases; al; al = al->next)
	fprintf(f, "#alias %s=%s\n", al->name, al->subst);
    for(ac = actions; ac; ac = ac->next) {
	escape_specials(buf, ac->pattern, "\\{}\"");
	fprintf(f, "#action %c%s \"%s\"=%s\n", (ac->active < 0) ? '-' : '+',
		ac->label, buf, ac->commands);
    }
    for(m = markers; m; m = m->next) {
	char *p = buf;
	if(m->pattern[0] == ' ') *p++ = '\\';
	escape_specials(p, m->pattern, "\\{}=");
	fprintf(f, "#mark %s=%s\n", buf, attr_name(m->attrcode));
    }
    /* NOTE: deleted default keys should be saved too, as "#bind <key>=" */
    for(k = keydefs; k; k = k->next)
	if(k->type == KEY_NORMAL) {
	    escape_seq(buf, k->sequence);
	    fprintf(f, "#bind %s %s=%s\n", k->label, buf, k->keydef.str);
	}
}

#ifndef OMIT_SESSIONS
/*
 * Below are multiple-session support functions.
 *
 * NOTE: sess_active[0] holds total no. of connections
 */

/*
 * return session's index or socket descriptor given id,
 * or 0 if invalid id is given
 */
int find_sess(id, notify)
char *id;
int notify;
{
    int i, n;
    for(i = main_sockfd, n = sess_active[0]; n; i++)
	if(sess_id[i]) {
	    n--;
	    if(!strcmp(sess_id[i], id))
		return i;
	}
    if(notify)
	printf("#no such connection\n");
    return 0;
}

/*
 * send text to specified session
 */
void send_to_sess(id, str, len)
char *id, *str; int len;
{
    int t = find_sess(id, 1);
    if(t) {
	int backfd = sockfd;
	sockfd = t;
	send_data(str, len);
	sockfd = backfd;
    }
}

/*
 * connect another session
 */
int connect_sess(id, host, port)
char *id, *host, *port;
{
    int fd;
    if((!host && !(host = hostname)) || (!port && !(port = portnumber))) {
	printf("#please specify both hostname and port number.\n");
	return -1;
    }
    /* dial the number by moving the right index in small circles */
    if((fd = do_connect(host, port)) < 0)
	return fd;
    FD_SET(fd, &fdset);		/* add socket to select() set */
    sess_active[0]++;
    if(!main_sockfd)		/* save main socket fd */
	main_sockfd = sockfd = fd;
    sess_active[fd] = 1;	/* use "#snoop" to toggle this */
    sess_id[fd] = str_dup(id);
    return fd;
}

/*
 * disconnect session
 */
void disconnect_sess(sfd)
int sfd;
{
    do_disconnect(sfd);
    printf("#connection \"%s\" closed.\n", sess_id[sfd]);
    if(sfd == main_sockfd)		/* main connection closed */
	exitcancan("");
    FD_CLR(sfd, &fdset);
    sess_active[0]--;
    sess_active[sfd] = 0;
    str_assign(sess_id + sfd, (char *)0);
}

/*
 * toggle output display from another session
 */
void snoop_sess(id)
char *id;
{
    int sfd = find_sess(id, 1);
    if(sfd)
	sess_active[sfd] ^= 1;
}

/*
 * show open connections
 */
void show_connections()
{
    int i, n = sess_active[0];
    if(n) {
	sprintf(insertbuf, "#The following connection%s open:",
		(n == 1) ? " is" : "s are");
	for(i = main_sockfd; to_pager(insertbuf), --n >= 0; i++) {
	    while(!sess_id[i]) i++;
	    sprintf(insertbuf, "##%-8s%s", sess_id[i],
		    !sess_active[i] ? "" :
		    i == main_sockfd ?  " (active)" : " (snooped)");
	}
    } else
	printf("#No open connections.\n");
}
#endif

