/*
 * cancan.h
 * various constants etc
 */

/* CONSTANTS */

#define DEFAULT_EDITOR	"emacs"	/* (environment) variable EDITOR overrides */
#define DEFAULT_PAGER	"more"	/* (environment) variable PAGER overrides */

#define CMDSEP		';'	/* command separator character */
#define ESCAPE_CHAR	'\035'	/* restore proper linemode */

#define LOOP_LIMIT	300	/* max number of loop iterations */

#define MAXMAPLEN	512	/* maximum length of automapped path */
#define WORDSKEPT	256	/* number of words kept for TAB-completion */
#define MAXHIST		128	/* number of history lines kept */

#define FLASHDELAY   500000	/* time of parentheses flash in microsecs */
#define KBD_TIMEOUT   35000	/* timeout for keyboard read in microsecs */
#define SCH_HZ		100	/* scheduled commands delay precision */

#define MAXFDSCAN	 32	/* max number of fd to scan in select() */
#define CAPLEN		 20	/* max length of a terminal capability */
#define LINELEN	       1024	/* maximum line length */
#define BUFSIZE	       2048	/* generic i/o buffer size */

/* attention flags: */
#define ATTN_FLASHBACK	  1	/* move cursor back after flashparen() */
#define ATTN_FOOTNOTE	  2	/* show/clear the bottom line footnote */
#define ATTN_WINCHSIG	  4	/* keep up with window size changes */
#define ATTN_CHILDSIG	  8	/* finish editing sessions */

/* linemode values: */
#define LM_NORMAL	  0
#define LM_CHAR		  1
#define LM_ESCAPED	  2

/* constants passed to input handlers: */
#define DO_CLEAR_INPUT	 -1
#define DO_REDRAW_INPUT	 -2
#define DO_CLEAR_PROMPT	 -3
#define DO_REDRAW_PROMPT -4


extern char **environ;
extern int errno;
#if !defined(USE_STRERROR)
extern int sys_nerr;
extern char *sys_errlist[];
#endif

#if !defined(__STDC__) && !defined(const)
#define const
#endif

#if defined(USE_BCOPY) && !defined(memcpy)
#define memcpy(a, b, c) bcopy((b), (a), (c))
#endif

#if MAXFDSCAN > 32
#define COPY_FDS(src, dst) memcpy((char *)(dst), (char *)(src), sizeof(fd_set))
#else
#define COPY_FDS(src, dst) *((long *)(dst)) = *((long *)(src))
#endif


/* TYPES & STRUCTURES */

enum opt_token {
    OPT_AUTOSAVE, OPT_BACKUPFILE, OPT_FLASHPAREN, OPT_PAGERWRAP,
    OPT_MAZEMAPPING, OPT_SPEEDWALK, OPT_VERBATIM, OPT_WORDWRAP,
    OPT_ERROR
};

#define OPTION_NAMES {						\
    "autosave", "backupfile", "flashparen", "pagerwrap", 	\
    "mazemapping", "speedwalk", "verbatim", "wordwrap",		\
    (char *)0							\
}

/* make sure all options have initial values here */
#define OPTION_DEFAULTS {					\
    1, 1, 1, 1,							\
    0, 0, 0, 1,							\
}

typedef long expr_value;

typedef int (input_proc)();
typedef void (cmd_func)();
typedef void (editfunc)();

struct defbind_node {
    char label[7], type;
    char *seq, *name;
    editfunc *func;
};

typedef struct {
    int pos, size;
    char buf[LINELEN];
} lbuf;

#define LBUF_ACCUM(lb, c) \
    ( (lb)->size > (lb)->pos ? (lb)->buf[(lb)->pos++] = (c), 1 : 0 )


/* SHARED FUNCTIONS */

/* from cancan.c */
void printver();
void exitcancan();
int get_one_char();
int execute();
void intrsig();
void suspend_cancan();
void syswarn();
void syserr();
void save_def_file();
void to_footnote();
void to_capfile();
int find_sess();
void send_to_sess();
int connect_sess();
void disconnect_sess();
void snoop_sess();
void show_connections();

/* from util.c */
int do_fork();
void do_execve();
FILE *do_popen();
int do_pclose();
lbuf *lbuf_alloc();
void lbuf_free();
int lbuf_accum();
int str_len();
int str_copy();
char *str_dup();
char *strn_dup();
void str_assign();
char *get_env();
char *get_envn();
void put_env();
void put_envn();
void init_env();

/* from edit.c */
void line_edit();
int handle_echoless();
int handle_charmode();
int handle_escaped();
void escape_to_linemode();
void set_input_handler();
void put_word();
void put_word_permanent();
void init_completions();

/* from expr.c */
char *expr_calc();
char *expr_getval();
char *expr_getnum();
char *expr_val2str();
int str_atoi();

/* from mapping.c */
void remove_steps();
void retrace_steps();
void show_automap();
void add_dir_to_map();

/* from tel.c */
int do_connect();
void do_disconnect();
int tn_process();
void set_linemode();
void send_term_size();
void send_to_host();
#define send_data(buf, len) send_to_host(buf, len, 0)
#define send_line(buf, len) send_to_host(buf, len, 1)
void flush_output_buffer();
void write_socket();
int read_socket();

/* SHARED VARS */

/* from cancan.c */
extern FILE *capfile;
extern char *initstr;

extern int main_sockfd;
extern int sockfd;		/* socket file descriptor */
extern int attention;		/* attn flags, should be sig_atomic_t */

extern int mp_active;
extern int new_settings;	/* action/alias/etc has changed */
extern int verbose;		/* this is set if cmd is user-typed */
extern int expect_prompt;	/* set when we send_to_host */

extern char current_options[];	/* various option flags, see above */
#define OPTION(x) current_options[x]

extern int linemode;

extern int flashback;		/* cursor is on excursion, put back there */
extern char *footnote;
extern char *prefixstr;
extern char insertbuf[];
extern char promptstr[];	/* prompt string */

/* from edit.c */
extern struct defbind_node const defbinds[];
extern input_proc *use_handler;
extern input_proc *input_handler;
extern int pos;			/* cursor position in edbuf */
extern int edlen;		/* characters in edbuf */
extern char edbuf[];		/* input line buffer */


#if defined(TELNETBUG) && !defined(WITH_STDCOLOUR)
#define WITH_STDCOLOUR
#endif

#if defined(TERM) && !defined(OMIT_SESSIONS)
#define OMIT_SESSIONS
#endif

#ifndef OMIT_SESSIONS
#define is_main_sess() (sockfd == main_sockfd)
#else
#define is_main_sess() (1)
#endif

