/*
 * cmd.c
 * functions for cancan's built-in #commands
 *
 * (created: Finn Arne Gangstad, Dec 25th, 1993)
 */

#include <stdio.h>
#include <string.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "ctypes.h"
#include "cmd.h"
#include "beam.h"

static void show_aliases();
static void show_actions();
static void show_keybinds();
static void keybind_modify();
static int key_collision();
static char *read_seq();
static void show_marks();
static int parse_attributes();
#ifndef OMIT_SCHEDULE
static void show_schedule();
static void parse_schedule();
static long parse_delay();
#endif
static int toggle_option();

static void
    cmd_action(), cmd_add(), cmd_alias(), cmd_bind(), cmd_cancel(),
    cmd_capture(), cmd_echo(), cmd_edit(), cmd_for(), cmd_help(),
    cmd_hilite(), cmd_identify(), cmd_init(), cmd_map(), cmd_mark(),
    cmd_option(), cmd_prefix(), cmd_quote(), cmd_read(), cmd_record(),
    cmd_retrace(), cmd_set(), cmd_speedwalk(), cmd_switch(), cmd_ver(),
    cmd_while(), cmd_wrap();
#ifndef OMIT_SESSIONS
static void cmd_connect(), cmd_snoop(), cmd_zap();
#endif
#ifndef OMIT_SCHEDULE
static void cmd_schedule();
#endif
#ifndef NOSHELL
static void cmd_shell();
#endif
#ifdef WITH_STDCOLOUR
static void cmd_stdcolour();
#endif

struct cmd_def {
    char *name;		/* command name */
    char *help;		/* short help */
    cmd_func *func;	/* function to call */
};

static struct cmd_def const commands[] = {
    { "#help", "[topic]\t\t\thelp on commands, or this text", cmd_help },
    { "#17", "command\t\t\trepeat 'command' 17 times", 0 },
#ifndef NOSHELL
    { "#!", "shell-command\t\texecute a shell command using /bin/sh",
	cmd_shell },
#endif
    { "#action", "[name][\"pattern\"][=cmd]\tlist/delete/define actions",
	cmd_action },
    { "#add", "[word]\t\t\tadd words to word completion list", cmd_add },
    { "#alias", "[name[=text]]\t\tlist/delete/define aliases", cmd_alias },
    { "#bind", "[name[=command]]\t\tlist/delete/define key bindings",
	cmd_bind },
    { "#cancel", "[number]\t\tcancel editing session", cmd_cancel },
    { "#capture", "[filename]\t\tbegin/end of capture to file", cmd_capture },
#ifndef OMIT_SESSIONS
    { "#connect", "[id[ init][=address]]\tconnect a new session", cmd_connect },
#endif
    { "#echo", "[-n ][text]\t\tdisplay text on screen", cmd_echo },
    { "#edit", "\t\t\t\tlist editing sessions", cmd_edit },
    { "#for", "a b [c] cmd\t\titerate a command", cmd_for },
    { "#hilite", "[attr]\t\t\thighlight your input line", cmd_hilite },
    { "#identify", "[startact endact]\tsend MUME client identification",
	cmd_identify },
    { "#init", "[-][[+]string]\t\tsend string to host on connect", cmd_init },
    { "#map", "[[-]number]\t\tshow/clear (auto)map", cmd_map },
    { "#mark", "[string[=attr]]\t\tlist/define markers", cmd_mark },
    { "#option", "[[+|-]option ...]\tset your preferences", cmd_option },
    { "#prefix", "string\t\t\tprefix all lines with string", cmd_prefix },
    { "#quote", "[on|off]\t\t\ttoggle verbatim-flag", cmd_quote },
    { "#read", "file\t\t\tread lines from file/pipe and send", cmd_read },
    { "#record", "[name]\t\t\trecord typed lines into alias", cmd_record },
    { "#retrace", "[[-]number]\t\tretrace the last number steps", cmd_retrace },
#ifndef OMIT_SCHEDULE
    { "#schedule", "name[ delay][=value]\tlist/delete/define delayed commands",
	cmd_schedule },
#endif
    { "#set", "name[=value]\t\tset variable value", cmd_set },
#ifndef OMIT_SESSIONS
    { "#snoop", "id\t\t\ttoggle output display for session", cmd_snoop },
#endif
    { "#speedwalk", "[on|off]\t\ttoggle speedwalk mode", cmd_speedwalk },
#ifdef WITH_STDCOLOUR
    { "#stdcolour", "[attr]\t\tset standard background colour", cmd_stdcolour },
#endif
    { "#switch", "expression command ...\tswitch statement", cmd_switch },
    { "#ver", "\t\t\t\tshow cancan version", cmd_ver },
    { "#while", "expression command\titerate a command", cmd_while },
    { "#wrap", "[on|off]\t\t\tturn word wrapping on/off", cmd_wrap },
#ifndef OMIT_SESSIONS
    { "#zap", "id\t\t\t\tdisconnect a session", cmd_zap },
#endif
    { 0, 0, 0 }
};

static char * const option_names[] = OPTION_NAMES;
static char * const attr_names[] = ATTR_NAMES;
#define get_attr_token(p) ((attr_token)get_token(p, attr_names))

/*
 * put commands into cmdhash and word completion
 */
void init_commands()
{
    const struct cmd_def *p = commands;
    aliasnode *hn;
    do if(p->func) {
	int len = str_len(p->name);
	hn = new_aliasnode(p->name, len);
	hn->func = p->func;
	hash_insert_(cmdhash, hn);
	if(len >= 3)
	    put_word_permanent(p->name, len);
    } while((++p)->name);
}

static void cmd_action(arg)
char *arg;
{
    if(!*arg)
	show_actions();
    else
	parse_action(arg);
}

/*
 * show defined actions
 */
static void show_actions()
{
    actionnode *p = actions;
    if(p) {
	sprintf(insertbuf, "#The following action%s defined:",
		!p->next ? " is" : "s are");
	for( ; to_pager(insertbuf), p; p = p->next)
	    sprintf(insertbuf, "#action %c%s \"%s\"=%s",
		    (p->active < 0) ? '-' : '+',
		    p->label, p->pattern, p->commands);
    } else
	printf("#No actions are defined.\n");
}

/*
 * parse the #action command
 */
void parse_action(str)
char *str;
{
    actionnode **np;
    char *label = 0, *pattern = 0, *rest = 0;
    char *s = str, sign = 0;

    if(*s != '"') {				/* the label */
	if(*s == '-' || *s == '+') sign = *s++;
	s = chop_label((label = s));
	if(s == label)
	    { printf("#action: invalid label\n"); return; }
    }
    if(*s == '"') {				/* the pattern */
	s = chop_arg((pattern = s + 1), '"');
	if(!s)
	    { printf("#action: pattern not \"-terminated\n"); return; }
	unescape(pattern);
	s = skipspace(s);
    }
    if(*s) {					/* the rest */
	if(*s != '=')
	    { printf("#action: incorrect syntax\n"); return; }
	*s = 0; rest = s + 1;
	/* if(!brace_balance(rest)) return; */
    }
    /*
     * at this point, either label or pattern or both are set
     * NOTE: we do not check if two actions have the same pattern.
     * Should we warn then? Deactivate the second one? hmm.
     */

    np = label ? lookup_action(label) : lookup_action_pattern(pattern);
    if(*np) {
	if(!rest && !sign && !pattern) {
	    rest = insertbuf;
	    sprintf(rest, "#action %s%s \"",
		    ((*np)->active < 0) ? "-" : "", (*np)->label);
	    rest += str_len(rest);
	    /* ';' here should actually be CMDSEP */
	    rest += escape_specials(rest, (*np)->pattern, "\\{};\"");
	    sprintf(rest, "\"=%s", (*np)->commands);
	} else if(rest && !*rest) {
	    if(verbose)
		printf("#action %s \"%s\" deleted\n",
			(*np)->label, (*np)->pattern);
	    delete_actionnode(np);
	    new_settings = 1;
	} else {
	    if(sign && ((*np)->active < 0) != (sign == '-')) {
		(*np)->active = -1 - (*np)->active;
	    } else
		sign = 0;
	    if(pattern && strcmp((*np)->pattern, pattern)) {
		str_assign(&(*np)->pattern, pattern);
		study_action_pattern(*np);
		sign = 1;
	    }
	    if(rest && strcmp((*np)->commands, rest)) {
		str_assign(&(*np)->commands, rest);
		sign = 1;
	    }
	    if(sign == 1) {
		if(verbose)
		    printf("#changed action: %s%s \"%s\"=%s\n",
			    ((*np)->active < 0) ? "-" : "", (*np)->label,
			    (*np)->pattern, (*np)->commands);
		new_settings = 1;
	    } else if(sign) {
		if(verbose)
		    printf("#action %s \"%s\" is now %s\n", (*np)->label,
			    (*np)->pattern, (sign != '-') ? "on" : "off");
	    }
	}
    } else if(!rest || !*rest || !pattern) {
	if(label)
	    printf("#no action labelled '%s'\n", label);
	else
	    printf("#no action with pattern \"%s\"\n", pattern);
    } else {
	static int last;
	static char buf[16];
	if(!label) {
	    label = buf;
	    do sprintf(label, "%d", ++last); while(*lookup_action(label));
	    /* reset the counter if next value will be a power of 10? */
	}
	np = add_actionnode(label, pattern, rest);
	if(sign == '-')
	    (*np)->active = -1;
	study_action_pattern(*np);
	if(verbose)
	    printf("#new action: %s \"%s\"=%s\n", label, pattern, rest);
	new_settings = 1;
    }
}

/*ARGSUSED*/
static void cmd_add(arg, ctx)
char *arg;
struct parse_context *ctx;
{
    /* NOTE: should we check that param 0 is a single word? */
    struct param *p = ctx->params;
    if(p && p[0].len >= 3)
	put_word(p[0].s, p[0].len);
}

static void cmd_alias(arg)
char *arg;
{
    if(!*arg)
	show_aliases();
    else
	parse_alias(arg);
}

/*
 * show defined aliases
 */
static void show_aliases()
{
    aliasnode *p = aliases;
    if(p) {
	sprintf(insertbuf, "#The following alias%s defined:",
		!p->next ? " is" : "es are");
	for( ; to_pager(insertbuf), p; p = p->next)
	    sprintf(insertbuf, "#alias %s=%s", p->name, p->subst);
    } else
	printf("#No aliases are defined.\n");
}

/*
 * parse the #alias command
 */
void parse_alias(str)
char *str;
{
    aliasnode **np;
    char *name = str;
    char *rest = chop_label(str);
    if(rest == name)
	{ printf("#invalid alias name\n"); return; }
    if(*rest && *rest != '=')
	{ printf("#missing a '=' in alias definition\n"); return; }
    if(*rest) { *rest++ = 0; } else { rest = 0; }
    np = lookup_alias(name);
    if(*np) {
	if(!rest)
	    sprintf(insertbuf, "#alias %s=%s", name, (*np)->subst);
	else if(!*rest) {
	    if(verbose)
		printf("#deleting alias: %s=%s\n", name, (*np)->subst);
	    delete_aliasnode(np);
	    new_settings = 1;
	} else if(strcmp(rest, (*np)->subst)) {
	    /* if(!brace_balance(rest)) return; */
	    str_assign(&(*np)->subst, rest);
	    if(verbose)
		printf("#changed alias: %s=%s\n", name, rest);
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	printf("#unknown alias: %s\n", name);
    } else if(!strcmp(name, "#alias")) {
	printf("#won't alias '%s'\n", name);
    } else {
	/* if(!brace_balance(rest)) return; */
	np = add_aliasnode(name, rest);
	if(verbose)
	    printf("#new alias: %s=%s\n", name, rest);
	new_settings = 1;
    }
}

static void cmd_bind(arg)
char *arg;
{
    if(!*arg)
	show_keybinds();
    else
	parse_bind(arg);
}

/*
 * list keyboard bindings
 */
static void show_keybinds()
{
    keynode *p;
    int count = 0;
    for(p = keydefs; p; p = p->next)
	if(p->type == KEY_NORMAL) {
	    if(!count++)
		to_pager("#User-defined keys:");
	    sprintf(insertbuf, "#bind %s=%s", p->label, p->keydef.str);
	    to_pager(insertbuf);
	}
    if(!count)
	printf("#No user-defined key bindings.\n");
}

/*
 * parse the #bind command
 */
void parse_bind(str)
char *str;
{
    keynode **np;
    char *label = str;
    char *rest = chop_arg(str, '=');
    char *kseq = chop_label(label);
    if(kseq == label)
	{ printf("#bind: invalid key label\n"); return; }
    if(*kseq) { unescape_seq(kseq); } else { kseq = 0; }
    np = lookup_key(label);
    if(*np) {
	char *key_cmd = (*np)->keydef.str;
	if((*np)->type == KEY_PROTECTED)
	    printf("#key '%s' is predefined and cannot be modified.\n", label);
	else if(kseq && strcmp(kseq, (*np)->sequence))
	    printf("#key '%s' already bound; will not rebind.\n", label);
	else if(!rest)
	    sprintf(insertbuf, "#bind %s=%s", label, key_cmd);
	else if(!*rest) {
	    if(verbose)
		printf("#key binding deleted: %s=%s\n", label, key_cmd);
	    delete_keynode(np);
	    new_settings = 1;
	} else if(strcmp(rest, key_cmd)) {
	    /* if(!brace_balance(rest)) return; */
	    str_assign(&(*np)->keydef.str, rest);
	    keybind_modify(*np);
	    if(verbose)
		printf("#redefined key: %s=%s\n", label, rest);
	    new_settings = 1;
	}
    } else if(!rest || !*rest) {
	/* NOTE: might let the command default to &nop perhaps? */
	printf("#no such key: '%s'\n", label);
    } else if(kseq ? !key_collision(kseq, 1) :
	    !key_collision(kseq = read_seq(), 0)) {
	/* NOTE: we should not allow read_seq in every situation */
	/* if(!brace_balance(rest)) return; */
	np = add_keynode(label, kseq, rest);
	keybind_modify(*np);
	if(verbose)
	    printf("#new key binding: %s=%s\n", label, rest);
	new_settings = 1;
    }
}

/*
 * determine the type of keybind, and assign keydef.func.
 */
static void keybind_modify(node)
keynode *node;
{
    const struct defbind_node *p;
    char *key_cmd = node->keydef.str;
    node->keydef.func = run_command;
    node->type = KEY_NORMAL;
    if(*key_cmd == '&')
	for(p = defbinds; p->label[0]; p++) {
	    if(!p->name || strcmp(p->name, key_cmd)) continue;
	    node->keydef.func = p->func;
	    if(strcmp(p->label, node->label)) continue;
	    if(!p->seq || strcmp(p->seq, node->sequence)) continue;
	    node->type = p->type; break;
	}
}

/*
 * determine if a keysequence collides with that of another key.
 */
static int key_collision(keyseq, shadow)
char *keyseq;
int shadow;
{
    keynode **np;
    if(!*keyseq || isprint(*keyseq)) {
	printf("#that is not a redefinable key.\n");
	return 1;
    }
    /*
     * NOTE: tree lookup would be natural here, but
     * we need a (keynode **) to delete it.
     */
    for(np = &keydefs; *np; np = &(*np)->next) {
	register char *a = (*np)->sequence;
	register char *b = keyseq;
	do if(!*a || !*b) {
	    if(*a || *b) {
		printf("#colliding keysequence: key '%s'\n", (*np)->label);
	    } else if((*np)->type == KEY_DEFAULT && shadow) {
		delete_keynode(np); return 0;
	    } else if((*np)->type == KEY_PROTECTED) {
		printf("#key '%s' is predefined and cannot be modified.\n",
			(*np)->label);
	    } else
		printf("#key already bound as: %s=%s\n",
			(*np)->label, (*np)->keydef.str);
	    return 1;
	} while(*a++ == *b++);
    }
    return 0;
}

/*
 * prompt the user to press a key, read an escape sequence from the
 * keyboard, print it. return static pointer to sequence.
 *
 * NOTE: to allow the user to set "meta keys" when the terminal has
 * no meta mode, there's no timeout after the leading escape.
 * NOTE: NUL's in key sequences are converted to \200's.
 */
static char *read_seq()
{
    static char kbuf[CAPLEN];
    int i, timeout;
    char c, *s;

    printf("#please press the key you want to define:");
    fflush(stdout);

    for(timeout = 0, i = 0; i < CAPLEN - 1; i++) {
	if(!get_one_char(kbuf + i, timeout)) break;
	c = kbuf[i];
#if !defined(metakey) || metakey
	if(metakey && (c & 0200)) {
	    if(!i) kbuf[i++] = '\033';	/* prepend an escape */
	    kbuf[i] = (c & 0177);	/* clear the meta bit */
	}
#endif
	if(i || c != '\033') timeout = 1;
    }
    kbuf[i] = 0;

    for(s = kbuf; i; s++, i--)
	if(*s == '\033') {
	    printf(" esc");
	} else if(*s == ' ') {
	    printf(" space");
	} else if((unsigned char)*s < ' ') {
	    printf(" ^%c", *s | 0100);
	    if(!*s) *s = '\200';
	} else if(!isprint(*s)) {
	    printf(" \\%03o", (int)*s);
	} else
	    printf(" %c", *s);
    newline();
    return kbuf;
}

static void cmd_cancel(arg)
char *arg;
{
    editsess *sp = edit_sess;
    int num;
    if(!sp) {
	printf("#No active sessions to cancel.\n");
    } else if(*arg) {
	if((num = str_atoi(arg)))
	    for( ; sp; sp = sp->next)
		if(num == sp->num) {
		    cancel_edit(sp);
		    return;
		}
	printf("#no such session - use the session number\n");
    } else if(sp->next) {
	printf("#several sessions active, use #cancel <number>\n");
    } else
	cancel_edit(sp);
}

static void cmd_capture(arg)
char *arg;
{
    if(capfile) {
	if(*arg)
	    printf("#capture already active.\n");
	else {
	    fclose(capfile);  capfile = 0;
	    printf("#end of capture to file.\n");
	}
    } else {
	char *mode = "w";
	if(*arg == '>') {
	    arg = skipspace(arg + 1);
	    mode = "a";
	}
	if(!*arg)
	    printf("#capture to what file?\n");
	else if(!(capfile = fopen(arg, mode)))
	    printf("#error writing file %s\n", arg);
	else
	    printf("#capture to %s begins.\n", arg);
    }
}

static void cmd_echo(arg)
char *arg;
{
    int nl = 1;
    if(arg[0] == '-' && arg[1] == 'n') {		/* #echo -n ... */
	if(!arg[2]) return;
	if(arg[2] == ' ')
	    { arg += 3; nl = 0; }
    }
    fwrite(arg, 1, unescape_seq(arg), stdout);
    if(nl) newline();
}

/*ARGSUSED*/
static void cmd_edit(arg)
char *arg;
{
    editsess *s = edit_sess;
    if(s) {
	do if(s->state) {
	    sprintf(insertbuf, "#%s session %d '%s'",
		    s->mode == 'E' ? "editing" :
		    s->mode == 'V' ? "viewing" : "cancelled",
		    s->num, s->desc);
	    to_pager(insertbuf);
	} while((s = s->next));
    } else
	printf("#No active editors.\n");
}

/*
 * NOTE: we could use expr_getval() in #for command, and let the begin,
 * end, step be any expressions (including numbers). Then we should insist
 * that the command be enclosed in {}, though. hmm.
#define expr_getnum(s, pv) expr_getval(s, pv)
 */

static void cmd_for(arg, ctx)
char *arg;
struct parse_context *ctx;
{
    /* NOTE: #for command (and others) should be interruptable */
    char buf[4 * sizeof(expr_value) + 4], *s = arg;
    expr_value begin, end, step, i;
    int llim;
    if((s = expr_getnum(s, &begin)) && (s = expr_getnum(s, &end))) {
	char *cmd = expr_getnum(s, &step);
	if(!cmd) { cmd = s; step = 1; }
	buf[0] = 'I'; buf[1] = '=';
	for(i = begin, llim = LOOP_LIMIT; llim; i += step, llim--) {
	    if(step >= 0 && i > end) break;
	    if(step <= 0 && i < end) break;
	    expr_val2str(buf + 2, i);
	    put_envn(buf, 1, 1);
	    parse(cmd, ctx->silent, (struct param *)0);
	}
	put_envn(buf, 1, 0);
    } else
	printf("#for: bad syntax\n");
}

static void cmd_hilite(arg)
char *arg;
{
    parse_hilite(arg);
}

void parse_hilite(str)
char *str;
{
    int attr = parse_attributes(str);
    if(attr >= 0) {
	/*
	 * NOTE: #hilite with !verbose (e.g. an action-bound command)
	 * could have a separate function.
	 */
	attr_string(edattr = attr, edattrbeg, edattrend);
	printf("#input highlighting is now %s%s%s.\n",
		edattrbeg, (edattr != NOATTRCODE) ? "on" : "off", edattrend);
	new_settings = 1;
    }
}

/*ARGSUSED*/
static void cmd_help(arg)
char *arg;
{
    const struct cmd_def *p;
    char *buf = insertbuf;
#ifdef HELPFILE
    if(*arg) {
	FILE *f = fopen(HELPFILE, "r");
	if(!f) {
	    printf("#can't open help file %s\n", HELPFILE);
	    return;
	}
	if(*arg == '#') arg++;
	while(fgets(buf, LINELEN, f)) if(buf[0] == '@') {
	    chop_str(buf + 1, '\n');
	    if(!buf[1] || !strcmp(arg, buf + 1)) {
		while(fgets(buf, LINELEN, f) && buf[0] != '@')
		    to_pager(buf);
		break;
	    }
	}
	insertbuf[0] = 0;	/* XXX */
	fclose(f);
	return;
    }
#endif
    to_pager("# commands available:");
    for(p = commands; p->name; p++) {
	sprintf(buf, "%s %s", p->name, p->help);
	to_pager(buf);
    }
}

static void cmd_identify(arg)
char *arg;
{
    char *p = *arg ? chop_arg(arg, ' ') : (arg = 0);
    str_assign(&edit_start, arg);
    str_assign(&edit_end, p);
    mp_active = 1;
    /* MUME doesn't prompt ident string */
    send_data(MCIDENT, MCIDENTLEN);
}

static void cmd_init(arg)
char *arg;
{
    if(!*arg) {
	if(initstr)
	    sprintf(insertbuf, "#init %s", initstr);
	else
	    printf("#init string %s.\n", "not defined");
    } else {
	if(arg[0] == '-' && !arg[1]) {
	    /* if(!initstr) return; */
	    arg = 0;
	    printf("#init string %s.\n", "cleared");
	} else if(!initstr || strcmp(arg, initstr)) {
	    /* if(!brace_balance(arg)) return; */
	    printf("#init string %s.\n", "changed");
	} else
	    return;
	str_assign(&initstr, arg);
	new_settings = 1;
    }
}

static void cmd_map(arg)
char *arg;
{
    if(arg[0] == '-' || isdigit(arg[0]))
	remove_steps(str_atoi(arg)); /* remove some steps without walking */
    else if(arg[0] == '+' && iswalkdir(arg[1]))
	add_dir_to_map(arg[1]);
    else
	show_automap();
}

static void cmd_mark(arg)
char *arg;
{
    if(!*arg)
	show_marks();
    else
	parse_marker(arg);
}

/*
 * show defined marks
 */
static void show_marks()
{
    marknode *p = markers;
    if(p) {
	sprintf(insertbuf, "#The following marker%s defined:",
		!p->next ? " is" : "s are");
	for( ; to_pager(insertbuf), p; p = p->next)
	    sprintf(insertbuf, "#mark %s=%s",
		    p->pattern, attr_name(p->attrcode));
    } else
	printf("#No markers are defined.\n");
}

/*
 * parse arguments to the #mark command
 */
void parse_marker(str)
char *str;
{
    int attrcode;
    marknode **np;
    char *pattern = str;
    char *rest = chop_arg(str, '=');
    /* if(rest) rest = skipspace(rest); */
    unescape(pattern);
    np = lookup_marker(pattern);
    if(*np) {
	if(!rest) {
	    rest = insertbuf;
	    rest += str_copy(rest, "#mark ");
	    if(*pattern == ' ') *rest++ = '\\';
	    /* ';' here should actually be CMDSEP */
	    rest += escape_specials(rest, pattern, "\\{};=");
	    *rest++ = '='; strcpy(rest, attr_name((*np)->attrcode));
	} else if(!*rest) {
	    if(verbose)
		printf("#mark deleted: %s=%s\n",
			(*np)->pattern, attr_name((*np)->attrcode));
	    delete_marknode(np);
	    new_settings = 1;
	} else if((attrcode = parse_attributes(rest)) >= 0 &&
		(*np)->attrcode != attrcode) {
	    (*np)->attrcode = attrcode;
	    if(verbose)
		printf("#changed mark: %s=%s\n",
			pattern, attr_name(attrcode));
	    new_settings = 1;
	}
    } else if(!rest || !*rest || !*pattern) {
	printf("#unknown marker: '%s'\n", pattern);
    } else if((attrcode = parse_attributes(rest)) >= 0) {
	np = add_marknode(pattern, attrcode);
	if(verbose)
	    printf("#new mark: %s=%s\n", pattern, attr_name(attrcode));
	new_settings = 1;
    }
}

/*
 * return a static pointer to name of given attribute code
 */
char *attr_name(attrcode)
int attrcode;
{
    static char name[32];	/* should be enough */
    attr_token fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);

    if(bg == ATTR_NONE)
	return attr_names[fg];
    if(fg < ATTR_COLOURS)
	sprintf(name, "%s on %s", attr_names[fg], attr_names[bg]);
    else
	sprintf(name, "on %s", attr_names[bg]);
    return name;
}

/*
 * parse attribute description in line.
 * Return attribute if successful, -1 otherwise.
 */
static int parse_attributes(line)
char *line;
{
    attr_token t, bg;
    if(*line) {
	for(t = bg = ATTR_NONE; ; ) {
	    if((t = get_attr_token(&line)) == ATTR_ERROR) break;
	    if(t == ATTR_FILL || (t < ATTR_COLOURS && line)) {
		if(t != ATTR_FILL && get_attr_token(&line) != ATTR_FILL) break;
		if(!line) break;
		if((bg = get_attr_token(&line)) >= ATTR_COLOURS) break;
	    }
	    if(line) break;
	    return ATTRCODE(t, bg);
	}
	printf("#invalid attribute syntax\n");
    }
    printf("Valid arguments are:\tnone, bold, underline, inverse\n");
    printf("\tor:\t\t[foreground] [on background]\nColours:");
    for(t = 0; t < ATTR_COLOURS; t++)
	printf((t < ATTR_COLOURS - 1) ? " %s," : " %s\n", attr_names[t]);
    return -1;
}

static void cmd_option(arg)
char *arg;
{
    if(!*arg)
	print_options(insertbuf);
    else
	parse_options(arg);
}

void parse_options(str)
char *str;
{
    char new[OPT_ERROR];
    int k;
    if(!strcmp(str, "save")) {
	save_def_file(1);
	return;
    }
    for(k = 0; k < OPT_ERROR; k++)
	new[k] = 0;
    for( ; ; ) {
	int n;
	char *s = str++;
	if(*s != '-' && *s != '+') break;
	k = get_token(&str, option_names);
	if(k == OPT_ERROR) break;
	new[k] = *s;
	if(str) continue;
	for(n = 0, k = 0; k < OPT_ERROR; k++) if(new[k]) {
	    char val = (new[k] != '-');
	    if(OPTION(k) == val) { new[k] = 0; continue; }
	    OPTION(k) = val;
	    if(verbose)
		printf((++n == 1 ? "#options changed: %c%s" : ", %c%s"),
			new[k], option_names[k]);
	}
	if(n) {
	    newline();
	    new_settings = 1;
	    if(new[OPT_AUTOSAVE] == '-')	/* user turns off autosave */
		save_def_file(0);
	}
	return;
    }
    printf("#syntax: #option [+option|-option] [...]\n");
    printf("Valid options:");
    for(k = 0; k < OPT_ERROR; k++)
	printf((k % 4) ? " %s" : "\n    %s", option_names[k]);
    newline();
}

void print_options(buf)
char *buf;
{
    int k;
    char *s = buf;
    s += str_copy(s, "#option");
    for(k = 0; k < OPT_ERROR; k++) {
	*s++ = ' ';
	*s++ = OPTION(k) ? '+' : '-';
	s += str_copy(s, option_names[k]);
    }
}

static void cmd_prefix(arg)
char *arg;
{
    if(!*arg) {
	printf("#prefix cleared.\n");
	arg = 0;
    }
    str_assign(&prefixstr, arg);
}

static void cmd_quote(arg)
char *arg;
{
    /*
     * TODO: "#quote <file". like verbatim mode, but after each enter
     * (or down key on the bottom of the history) a new input line is
     * read from the file.
     */
    int v = toggle_option(OPT_VERBATIM, arg);
    printf("#%s mode.\n", v ? "verbatim" : "normal");
}

static void cmd_read(arg)
char *arg;
{
    FILE *fp;
    char buf[LINELEN];
    if(*arg) {
	fp = (arg[0] == '|') ? do_popen(arg + 1, "r") : fopen(arg, "r");
	if(fp) {
	    /* assuming it's a unix text file */
	    while(fgets(buf, LINELEN, fp)) {
		int len = str_len(buf);
		if(len && buf[len - 1] == '\n') len--;
		send_line(buf, len);
	    }
	    if(arg[0] == '|') do_pclose(fp); else fclose(fp);
	} else
	    printf("#error opening %s\n", arg);
    } else
	printf("#syntax: #read filename or #read |command\n");
}

static void cmd_record(arg)
char *arg;
{
    aliasnode *r = recording_alias;
    int len;
    if(*arg) {
	if(r) {
	    printf("#already recording to alias '%s'\n", r->name);
	    return;
	}
	if(*chop_label(arg)) {
	    printf("#invalid alias name\n");
	    return;
	}
	len = str_len(arg);
	r = *lookup_cmdhash(arg, len);
	if(r) {
	    printf("#please choose another name for the alias\n");
	    return;
	}
	r = new_aliasnode(arg, len);
	hash_insert_(cmdhash, r);
	recording_alias = r;
	lock_aliasnode(r);
	if(verbose)
	    printf("#recording into alias '%s'\n", r->name);
    } else if(r) {
	if(verbose)
	    printf("#recording of alias '%s' stopped.\n", r->name);
	release_aliasnode(r);
	recording_alias = 0;
	new_settings = 1;
    } else
	printf("#not recording any alias\n");
}

static void cmd_retrace(arg)
char *arg;
{
    retrace_steps(str_atoi(arg));
}

#ifndef OMIT_SCHEDULE
static void cmd_schedule(arg)
char *arg;
{
    if(!*arg)
	show_schedule();
    else
	parse_schedule(arg);
}

/*
 * show scheduled commands
 */
static void show_schedule()
{
    schedulenode *p = schedule;
    if(p) {
	long delay = 0;
	sprintf(insertbuf, "#The following command%s scheduled:",
		!p->next ? " is" : "s are");
	for( ; to_pager(insertbuf), p; p = p->next) {
	    long t = (delay += p->delta) / SCH_HZ;
	    sprintf(insertbuf, "#schedule %-8s %2ld:%02d:%02d=%s", p->label,
		    t / 3600, (int)(t / 60 % 60), (int)(t % 60), p->commands);
	}
    } else
	printf("#No commands are scheduled.\n");
}

/*
 * parse the #schedule command
 */
static void parse_schedule(str)
char *str;
{
    long delay = 0;
    schedulenode **np;
    char *label = str;
    char *rest = chop_arg(str, '=');
    char *k = chop_label(label);
    if(k == label)
	{ printf("#schedule: invalid label\n"); return; }
    np = lookup_schedule(label);
    if(!*k && !*np && rest && *rest)	/* no label, delay=cmd */
	{ k = label; label = 0; }
    if(*k && !(delay = parse_delay(k)))
	{ printf("#schedule: a valid delay is expected\n"); return; }
    if(*np) {
	if(!rest && !delay) {
	    sprintf(insertbuf, "#schedule %s=%s", label, (*np)->commands);
	} else if(rest && !*rest) {
	    if(verbose)
		printf("#scheduled command deleted: %s=%s\n",
			label, (*np)->commands);
	    delete_schedulenode(np);
	} else {
	    if(rest && strcmp((*np)->commands, rest)) {
		str_assign(&(*np)->commands, rest);
		if(verbose)
		    printf("#scheduled command changed: %s=%s\n",
			    label, (*np)->commands);
	    } else if(delay) {
		if(verbose)
		    printf("#command rescheduled: %s=%s\n",
			    label, (*np)->commands);
	    }
	    if(delay)
		np = reschedule(np, delay);
	}
    } else if(!delay || !rest || !*rest) {
	printf("#no scheduled command with label '%s'\n", label);
    /* } else if(!brace_balance(rest)) { return; */
    } else {
	static int last;
	static char buf[16];
	if(!label) {
	    label = buf;
	    do sprintf(label, "_%d", ++last); while(*lookup_schedule(label));
	    /* reset the counter if next value is a power of 10? */
	}
	np = add_schedulenode(label, rest, delay);
	if(verbose)
	    printf("#new command scheduled: %s=%s\n", label, rest);
    }
}

/*
 * parse the command delay specification: [[hh:]mm:]ss
 * return the delay in units of 1/SCH_HZ of a second; 0 signifies error.
 * TODO: delays relative to day start "@HH:MM"?
 */
static long parse_delay(str)
char *str;
{
    char *s = str;
    long r = 0;
    int sign = 0;
    if(*s == '+') s++, sign = 1;
    while(*s)
	if(isdigit(*s))
	    { r += str_atoi(s); s += digit_span(s); }
	else if(*s == ':')
	    { r = r * 60; s++; }
	else
	    { r = 0; break; }
    r = r * SCH_HZ;
    return sign ? -r : r;
}
#endif

static void cmd_set(arg)
char *arg;
{
    if(*arg) {
	char *s = strchr(arg, '=');
	if(s) put_envn(arg, s - arg, (int)s[1]);
    } else {
	varnode *p;
	for(p = make_varlist(); p; p = p->next)
	    to_pager(p->name);
    }
}

#ifndef NOSHELL
static void cmd_shell(arg)
char *arg;
{
    char buf[32];
    reset_terminal();
    sprintf(buf, "error=%d", execute(arg));
    put_envn(buf, 5, 1);
    set_terminal();
    lastline();
}
#endif

static void cmd_speedwalk(arg)
char *arg;
{
    if(!parse_walk(arg)) {
	int v = toggle_option(OPT_SPEEDWALK, arg);
	printf("#speedwalk is now %s.\n", v ? "on" : "off");
    }
}

#ifdef WITH_STDCOLOUR
static void cmd_stdcolour(arg)
char *arg;
{
    int attrcode = parse_attributes(arg);
    if(attrcode >= 0) {
	if(set_stdcolour(attrcode)) {
	    printf("#standard colour %s.\n",
		    (attrcode == NOATTRCODE ? "cleared" : "set"));
	} else
	    printf("#please specify both foreground and background colours\n");
    }
}
#endif

static void cmd_switch(arg, ctx)
char *arg;
struct parse_context *ctx;
{
    expr_value val;
    char *rest = expr_getval(arg, &val);
    if(rest && (!*rest || *rest == ' ')) {
	do {
	    char *cmd = skipspace(rest);
	    if(!*cmd) break;
	    rest = chop_arg(cmd, ' ');
	    if(!val)
		{ parse(cmd, ctx->silent, (struct param *)0); break; }
	} while(--val, rest);
    } else
	printf("#switch: invalid expression\n");
}

/*ARGSUSED*/
static void cmd_ver(arg)
char *arg;
{
    printver();
}

static void cmd_while(arg, ctx)
char *arg;
struct parse_context *ctx;
{
    expr_value val;
    int llim;
    char *cmd = expr_getval(arg, &val);
    if(cmd && *cmd == ' ' && *(cmd = skipspace(cmd))) {
	for(llim = LOOP_LIMIT; llim && val; llim--) {
	    parse(cmd, ctx->silent, (struct param *)0);
	    expr_getval(arg, &val);
	}
    } else
	printf("#while: invalid syntax\n");
}

static void cmd_wrap(arg)
char *arg;
{
    int v = toggle_option(OPT_WORDWRAP, arg);
    printf("#word wrap %sabled.\n", v ? "en" : "dis");
}

#ifndef OMIT_SESSIONS
static void cmd_connect(arg)
char *arg;
{
    if(!*arg)
	show_connections();
    else {
	char *host = chop_arg(arg + 1, '=');
	if(*chop_label(arg)) {
	    printf("#connect: invalid session name\n");
	} else if(find_sess(arg, 0)) {
	    printf("#connect: session already exists\n");
	} else if(host && !*(host = skipspace(host))) {
	    printf("#connect: missing host address\n");
	} else {
	    char *port = 0;
	    if(host) port = chop_str(host, ' ');
	    connect_sess(arg, host, port);
	}
    }
}

static void cmd_snoop(arg)
char *arg;
{
    if(!*arg)
	printf("#snoop: no session name\n");
    else
	snoop_sess(arg);
}

static void cmd_zap(arg)
char *arg;
{
    int sfd;
    if(!*arg)
	printf("#zap: no session name\n");
    else if((sfd = find_sess(arg, 1)))
	disconnect_sess(sfd);
}
#endif

/*
 * change the option if arg is "on" or "off", return option state.
 * this may trash the arg!
 */
static int toggle_option(k, arg)
int k;
char *arg;
{
    static char * const on_off[] = { "off", "on", 0 };
    int v = get_token(&arg, on_off);
    if(v < 2 /* && !arg */)
	OPTION(k) = v;
    return OPTION(k);
}

