/*
 * edit.c
 * line editing functions for cancan
 */

#include <stdio.h>
#include <string.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "ctypes.h"

static void
    prev_char(), next_char(), prev_word(), next_word(),
    beginning_of_line(), end_of_line(), matching_bracket(),
    del_char_left(), del_char_right(),
    del_word_left(), del_word_right(), kill_to_eol(),
    transpose_chars(), transpose_words(),
    downcase_word(), upcase_word(), capitalize_word(),
    prev_line(), next_line(),
    beginning_of_history(), end_of_history(), enter_into_history(),
    complete_word(), complete_line(),
    redraw_input_line(), enter_line(), enter_line_verbatim();

#ifdef WITH_ISEARCH
static void reverse_i_search(), forward_i_search();
#endif

static void set_word_case();
static void run_edline();
static void swap_history();
static void flashparen();
static void put_in_history();
static void replace_edline();
static int skip_word();
static int skip_nonword();

/* circular word completion list */
static void delete_dup_words();
static void demote_word();
static int word_cmp();

#define WORD_UNIQUE	1	/* unique among preceding words */
#define WORD_RETAIN	2	/* permanent entry (a #command) */
#define WORD_CASE	4	/* use case sensitive matching	*/

static int wordindex;
static char *words[WORDSKEPT];
static char wordflags[WORDSKEPT];

#if (WORDSKEPT - 1) >> 8
static unsigned int wordprev[WORDSKEPT];	/* older word */
static unsigned int wordnext[WORDSKEPT];	/* newer word */
#else
static unsigned char wordprev[WORDSKEPT];
static unsigned char wordnext[WORDSKEPT];
#endif

/* history buffer */
static char *hist[MAXHIST];	/* saved history lines */
static int curline;		/* current history line */
static int pickline;		/* line to pick history from */

#if (MAXHIST - 1) & MAXHIST
#define HISTINDEX(i) ((unsigned)((i) + MAXHIST) % MAXHIST)
#else
#define HISTINDEX(i) ((i) & (MAXHIST - 1))
#endif

editfunc *last_edit_func;	/* last editor func that was used */
input_proc *use_handler;	/* alt. func to handle input */
input_proc *input_handler;	/* and the one currently active */
int pos;			/* cursor position in line */
int edlen;			/* length of current input line */
char edbuf[LINELEN];		/* line editing buffer */

/*
 * TODO:
 * C-o  operate-and-get-next
 * C-q  quoted-insert
 * M-?  possible-completions
 * &nop, &macro()
 * ( &denewbiefy, &enter-line-denewbiefied, if you care )
 */

struct defbind_node const defbinds[] = {
    { "enter",	2,  "\n",	"&enter-line",		enter_line },
    { "return",	2,  "\r",	"&enter-line",		enter_line },
    { "bs",	1,  "\b",	"&del-char-left",	del_char_left },
    { "del",	1,  "\177",	"&del-char-left",	del_char_left },
    { "tab",	1,  "\t",	"&complete-word",	complete_word },
    { "left",	1,  key_left,	"&prev-char",		prev_char },
    { "right",	1,  key_right,	"&next-char",		next_char },
    { "up",	1,  key_up,	"&prev-line",		prev_line },
    { "down",	1,  key_down,	"&next-line",		next_line },
    { "C-a",	1,  "\001",	"&beginning-of-line",	beginning_of_line },
    { "C-b",	1,  "\002",	"&prev-char",		prev_char },
    { "C-c",	2,  "\003",	"",			intrsig },
    { "C-d",	1,  "\004",	"&del-char-right",	del_char_right },
    { "C-e",	1,  "\005",	"&end-of-line",		end_of_line },
    { "C-f",	1,  "\006",	"&next-char",		next_char },
    { "C-g",	1,  "\007",	"&matching-bracket",	matching_bracket },
    { "C-k",	1,  "\013",	"&kill-to-eol",		kill_to_eol },
    { "C-l",	1,  "\014",	"&redraw-input-line",	redraw_input_line },
    { "C-n",	1,  "\016",	"&next-line",		next_line },
    { "C-p",	1,  "\020",	"&prev-line",		prev_line },
#ifdef WITH_ISEARCH
    { "C-r",	1,  "\022",	"&reverse-i-search",	reverse_i_search },
    { "C-s",	1,  "\023",	"&forward-i-search",	forward_i_search },
#endif
    { "C-t",	1,  "\024",	"&transpose-chars",	transpose_chars },
    { "C-y",	1,  "\031",	"&enter-into-history",	enter_into_history },
    { "C-z",	2,  "\032",	"",			suspend_cancan },
    { "M-ret",	1,  "\033\r",	"&enter-line-verbatim",	enter_line_verbatim },
    { "M-bs",	1,  "\033\b",	"&del-word-left",	del_word_left },
    { "M-del",	1,  "\033\177",	"&del-word-left",	del_word_left },
    { "M-tab",	1,  "\033\t",	"&complete-line",	complete_line },
    { "M-b",	1,  "\033b",	"&prev-word",		prev_word },
    { "M-c",	1,  "\033c",	"&capitalize-word",	capitalize_word },
    { "M-d",	1,  "\033d",	"&del-word-right",	del_word_right },
    { "M-f",	1,  "\033f",	"&next-word",		next_word },
    { "M-l",	1,  "\033l",	"&downcase-word",	downcase_word },
    { "M-t",	1,  "\033t",	"&transpose-words",	transpose_words },
    { "M-u",	1,  "\033u",	"&upcase-word",		upcase_word },
    { "M-<",	1,  "\033<",	"&beginning-of-history",beginning_of_history },
    { "M->",	1,  "\033>",	"&end-of-history",	end_of_history },
    { "",	0,  0,		0,			0 }
};

/*
 * the line editor. given some input chars (typically a single one),
 * add them to edit line, or call appropriate edit funcs.
 * TODO: pass <func, args> to (*input_handler)(), so that any key
 * could be recognized there?
 * TODO: insert multiple chars at a time if possible.
 */
void line_edit(s, len)
char *s; int len;
{
    static int nchars;			/* length of the (incomplete) seq. */
    static char typed[CAPLEN];		/* and the sequence itself */
    for( ; --len >= 0; s++) {
	register ktreenode *t;
	register int i;
	if(!nchars) {
	    char c;
	    if(input_handler && (*input_handler)((int)(unsigned char)*s))
		continue;
	    c = *s;
	    if(isprint(c)) {
		if(edlen < LINELEN - 1) {
		    insert_one_char(s);
		    last_edit_func = 0;
		    if(len) continue;
		    if(isrparen(c) && OPTION(OPT_FLASHPAREN))
			flashparen();
		    break;
		}
		continue;
#if !defined(metakey) || metakey
	    } else if((c & 0200) && metakey) {
		typed[0] = '\033';	/* prepend esc (meta prefix) */
		nchars = 1;
#endif
	    } else if((t = keymap[(unsigned char)c])) {
		/* assert(keymap[0] == keymap[0200]); */
		if(!t->cval) {
		    (*t->leaf.func)(t->leaf.str);
		    last_edit_func = t->leaf.func;
		} else {
		    typed[0] = c;
		    nchars = 1;
		}
		continue;
	    } else
		continue;
	}
	for(i = 1, t = keymap[(unsigned char)typed[0]]; ; ) {
	    char c = *s;
#if !defined(metakey) || metakey
	    if(metakey) c &= ~0200;	/* clear meta bit */
#endif
	    typed[nchars++] = c ? c : '\200';
	    while(t && t->cval) {	/* walk the tree */
		if(t->cval != typed[i])
		    { t = t->tree.right; continue; }
		t = t->tree.child; i++;
		if(i >= nchars) break;
	    }
	    if(i >= nchars && t) {
		if(t->cval) {		/* partial match */
		    if(!len) break;
		    s++; len--; continue;
		}
		(*t->leaf.func)(t->leaf.str);
		last_edit_func = t->leaf.func;
	    }
	    nchars = 0;
	    break;
	}
    }
}

/*
 * Flash cursor at parenthesis matching the one inserted before current pos
 */
static void flashparen()
{
    int lev = 0, i = pos;
    char c = 0;
    while(--i >= 0)
	if(isrparen(edbuf[i])) {
	    lev++;
	    if(!c) c = edbuf[i];
	} else if(islparen(edbuf[i]) && !--lev) {
	    if(LRPARENS(edbuf[i], c)) {
		attention |= ATTN_FLASHBACK;
		flashback = pos;
		moveto(i);
	    }
	    break;
	}
}

/*
 * handle input when sending password (line mode, no echo)
 * NOTE: we could draw asterisks here and handle del too.
 */
int handle_echoless(c)
int c;
{
    if(c == ESCAPE_CHAR) {
	escape_to_linemode();
    } else if(c >= 0) {
	/* if(c == '\003') intrsig(); */
        do {
	    if(c == '\r' || c == '\n') {
		newline();
		send_line(edbuf, edlen);
	    } else {
		edbuf[edlen++] = c;
		if(edlen < LINELEN) break;
		send_data(edbuf, edlen);
	    }
	    edlen = 0;
	} while(0);
	edbuf[pos = edlen] = 0;
    } else if(c == DO_CLEAR_PROMPT || c == DO_REDRAW_PROMPT)
	return 0;		/* don't override the prompt */
    return 1;
}

/*
 * handle input in char-by-char mode
 * never echo locally--it's almost always a mistake
 */
int handle_charmode(c)
int c;
{
    if(c == ESCAPE_CHAR) {
	escape_to_linemode();
    } else if(c >= 0) {
	char x = c;
	send_data(&x, 1);
    }
    return 1;
}

/*
 * handle user input in escaped mode
 */
int handle_escaped(c)
int c;
{
    if(expect_prompt)
	set_linemode(LM_NORMAL);
    return 0;
}

/* NOTE: we should maybe escape out of the (active) session */
void escape_to_linemode()
{
    set_linemode(LM_ESCAPED);
    newline();
    col0 = str_copy(promptstr, "cancan>");
    expect_prompt = 0;
    redraw_prompt();
}

void set_input_handler(func)
input_proc *func;
{
    use_handler = func;
    set_linemode(-1);
}

#ifdef WITH_ISEARCH
static void start_isearch();
static int isearch_slen;
static int isearch_sdir;

/*
 * handle incremental search
 * TODO: mark the matching substring in standout when redrawing edline.
 * Maybe handle del as well? Beep when match fails?
 */
static int handle_isearch(c)
int c;
{
    char buf[LINELEN], *h, *s, *sub;
    int i = pickline;
    if(c >= 0) {
	if(isprint(c)) {		/* hist[curline] can be 0 */
	    if(i == curline ? isearch_sdir > 0 : isearch_sdir < 0)
		i -= isearch_sdir;
	} else if(c == '\022') {	/* kludge! won't work if rebound */
	    isearch_sdir = -1;
	    c = 0;
	} else if(c == '\023') {	/* again */
	    isearch_sdir = 1;
	    c = 0;
	} else {			/* exit search if control key */
	    set_input_handler((input_proc *)0);
	    return 0;
	}
    }
    /* else if(c != DO_REDRAW_INPUT) return 0; */
    sprintf(buf, "#%s search history: `%.*s%c",
	    isearch_sdir < 0 ? "reverse" : "forward",
	    isearch_slen, edbuf + pos, (char)(c < 0 ? 0 : c));
    sub = buf + 26;
    if(c >= 0) for( ; ; ) {
	i = HISTINDEX(i + isearch_sdir);
	if(i == curline || !(h = hist[i])) break;
	if(!c && !strcmp(h, edbuf)) continue;
	if((s = strstr(h, sub))) {
	    int p = s - h;
	    swap_history(i);		/* recursion! */
	    moveto(p);
	    if(c) isearch_slen++;
	    break;
	}
    }
    strcpy(sub + isearch_slen, "'");
    to_footnote(buf);
    return (c >= 0);			/* don't override prompt/edline */
}

/*
 * perform reverse incremental search in history
 */
static void reverse_i_search()
{
    start_isearch(-1);
}

/*
 * perform forward incremental search in history
 */
static void forward_i_search()
{
    start_isearch(1);
}

static void start_isearch(dir)
int dir;
{
    isearch_slen = 0;
    isearch_sdir = dir;
    set_input_handler(handle_isearch);
    handle_isearch(DO_REDRAW_INPUT);	/* this will set footnote */
}
#endif


/*
 * move one char backward
 */
static void prev_char()
{
    if(pos)
	moveto(pos - 1);
}

/*
 * move one char forward
 */
static void next_char()
{
    if(pos < edlen)
	moveto(pos + 1);
}

/*
 * move one word forward
 */
static void next_word()
{
    moveto(skip_word(skip_nonword(pos)));
}

/*
 * move one word backward
 */
static void prev_word()
{
    int i = pos;
    if(i) {
	do --i; while(i && !isword(edbuf[i]));
	while(i && isword(edbuf[i - 1])) i--;
	moveto(i);
    }
}

/*
 * move cursor to beginning of line
 */
static void beginning_of_line()
{
    moveto(0);
}

/*
 * move cursor to end of line
 */
static void end_of_line()
{
    moveto(edlen);
}

/*
 * move to matching bracket
 */
static void matching_bracket()
{
    char *p, *brackets = "(){}[]<>";
    int i = pos;
    if((p = strchr(brackets, edbuf[i]))) {
	int nested = 0;
	int dir = (p - brackets) & 1 ? -1 : 1;
	while((i += dir) >= 0 && i < edlen)
	    if(edbuf[i] == p[0]) nested++;
	    else if(edbuf[i] == p[dir] && --nested < 0) {
		moveto(i);
		break;
	    }
    }
}


/*
 * delete a character to the right
 */
static void del_char_right()
{
    if(edlen > pos)
	delete_chars(1);
}

/*
 * delete a character to the left
 */
static void del_char_left()
{
    if(pos) {
	moveto(pos - 1);
	delete_chars(1);
    }
}

/*
 * delete word to the right
 */
static void del_word_right()
{
    int n = skip_word(skip_nonword(pos)) - pos;
    if(n) delete_chars(n);
}

/*
 * delete word to the left
 */
static void del_word_left()
{
    int n = pos;
    prev_word();
    if((n -= pos))
	delete_chars(n);
}

/*
 * erase everything to the end of line
 */
static void kill_to_eol()
{
    int n = edlen - pos;
    if(n) delete_chars(n);
}


/*
 * transpose the char under the cursor with previous one
 */
static void transpose_chars()
{
    if(edlen > 1) {
	char c1, c2;
	int k = pos;
	if(k) { if(k >= edlen) k--; k--; }
	moveto(k);
	c1 = edbuf[k]; c2 = edbuf[k + 1];
	replace_char(c2); replace_char(c1);
    }
}

/*
 * transpose the word behind the cursor with next one
 */
static void transpose_words()
{
    int a, b, c;
    prev_word();
    a = skip_word(skip_nonword(pos));
    b = skip_nonword(a);
    c = skip_word(b);
    if(b < c) {
	char buf[LINELEN], *s;
	sprintf(buf, "%.*s%.*s%.*s",
		c - b, edbuf + b, b - a, edbuf + a, a - pos, edbuf + pos);
	s = buf; do replace_char(*s); while(*++s);
    }
}

/*
 * set the word in lowercase, move to the end of it
 */
static void downcase_word()
{
    set_word_case(0);
}

/*
 * set the word in uppercase, move to the end of it
 */
static void upcase_word()
{
    set_word_case(LINELEN);
}

/*
 * capitalize a word, and move to the end of it
 */
static void capitalize_word()
{
    set_word_case(1);
}

/*
 * ncaps is the number of leading chars to put in uppercase.
 * rest of the word is downcased.
 */
static void set_word_case(ncaps)
int ncaps;
{
    int i = skip_nonword(pos);
    for(moveto(i); ; i++) {
	int c = (unsigned char)edbuf[i];
	if(!isword(c)) break;
	replace_char(--ncaps >= 0 ? toupper(c) : tolower(c));
    }
}

/*
 * return the position after leading isword() chars
 */
static int skip_word(i)
int i;
{
    while(isword(edbuf[i])) i++;
    return i;
}

/*
 * return the position after leading word separator chars
 */
static int skip_nonword(i)
int i;
{
    while(i < edlen && !isword(edbuf[i])) i++;
    return i;
}


/*
 * redraw the input line
 */
static void redraw_input_line()
{
    clear_all();
    col0 = 0;	/* nuke the prompt */
    redraw_input();
}

/*
 * enter the line as is
 */
static void enter_line_verbatim()
{
    run_edline(1);
}

/*
 * enter a line
 */
static void enter_line()
{
    int verbatim = 0;
    if(OPTION(OPT_VERBATIM) && edbuf[0] != '#' && edbuf[0] != '{')
	verbatim = 1;
    run_edline(verbatim);
}

static void run_edline(verbatim)
int verbatim;
{
    int len;
    to_capfile(edbuf, edlen, 1);
    skip_edline();
    insertbuf[0] = '\0';

    if(!verbatim) {				/* process the edline */
	new_settings = 0;
	verbose = 1;
	parse(edbuf, 1, (struct param *)0);
	check_line0();
	verbose = 0;
	if(new_settings && OPTION(OPT_AUTOSAVE))
	    save_def_file(0);
    } else
	send_line(edbuf, edlen);

    if(edlen > 1) {				/* commit to history */
	char *p = hist[HISTINDEX(curline - 1)];
	if(!p || strcmp(edbuf, p))
	    put_in_history(edbuf);
    }
    pickline = curline;

    edbuf[len = 0] = '\0';			/* and set new one */
    if(insertbuf[0])
	len = escape_command(edbuf, insertbuf);
    else if(prefixstr)
	len = str_copy(edbuf, prefixstr);
    edlen = pos = len;

    redraw_all();
}

/*
 * move the editline into history
 */
static void enter_into_history()
{
    if(edlen) {
	put_in_history(edbuf);
	pickline = curline;
	replace_edline("");
    }
}

/*
 * get previous line from history list
 */
static void prev_line()
{
    int i = HISTINDEX(pickline - 1);
    if(i != curline)
	swap_history(i);
}

/*
 * get next line from history list
 */
static void next_line()
{
    if(pickline != curline) {
	int i = HISTINDEX(pickline + 1);
	swap_history(i);
    }
}

/*
 * go to the beginning of history
 */
static void beginning_of_history()
{
    int i = curline;
    do {
	i = HISTINDEX(i + 1);
    } while(i != pickline && !hist[i]);
    swap_history(i);
}

/*
 * go to the end of history list
 */
static void end_of_history()
{
    swap_history(curline);
}

/*
 * swap editline with a history line
 */
static void swap_history(line)
int line;
{
    char *h;
    if(line != pickline && (h = hist[line])) {
	str_assign(hist + pickline, edbuf);
	pickline = line;
	replace_edline(h);
    }
}

/*
 * put a line in history at current position
 * and split its words into completion list.
 */
static void put_in_history(str)
char *str;
{
    int i = curline;
    char *s;
    str_assign(hist + i, str);
    curline = HISTINDEX(i + 1);
    for(s = str; ; str = ++s) {
	while(*s == '#') s++;
	while(isword(*s)) s++;
	if(s - str >= 3)	/* skip short words */
	    put_word(str, s - str);
	if(!*s) break;
    }
}

/*
 * clear old input line, copy new one into edbuf,
 * display it, set edlen and pos.
 */
static void replace_edline(str)
char *str;
{
    clear_input();
    edlen = pos = str_copy(edbuf, str);
    redraw_input();
}

/*
 * match backwards in history and complete entire lines
 */
static void complete_line()
{
    int i;
    char *h;
    if(edlen)
	for(i = curline; ; ) {
	    i = HISTINDEX(i - 1);
	    if(i == curline || !(h = hist[i])) break;
	    if(!strncmp(h, edbuf, edlen) /* && h[edlen] */)
		{ replace_edline(h); break; }
	}
}


/*
 * match and complete a word referring to the word list
 */
static void complete_word()
{
    static int curr_word, comp_len, root_len;
    char *root, *p;
    int k, n;

    /* determine curr_word, root, root_len */
    if(last_edit_func == complete_word && (k = comp_len)) {
	moveto(pos - k);
	n = pos - root_len;
    } else {
	for(n = pos; n && isword(edbuf[n - 1]); n--);
	while(n && edbuf[n - 1] == '#') n--;
	k = 0;  curr_word = wordindex;  root_len = pos - n;
    }
    root = edbuf + n;  comp_len = 0;

    /* look up next valid completion and print it */
    while((p = words[curr_word = wordprev[curr_word]])) {
	if(word_cmp(curr_word, root, root_len)) continue;
	if(!*(p += root_len)) continue;		/* already complete */
	if(edlen + (n = str_len(p)) - k >= LINELEN) continue; /* no room */
	delete_dup_words(curr_word);
	comp_len = n;
	/*
	 * NOTE: could change this to overtype the whole word (the
	 * beginning might have different capitalization).
	 */
	while(n && k)
	    { n--; k--; replace_char(*p++); }
	if(n)
	    { insert_chars(p, n); return; }
	break;
    }
    if(k) delete_chars(k);
}

/*
 * delete duplicate instances of a word
 */
static void delete_dup_words(i)
int i;
{
    int n, k;
    char *p;
    if(!(wordflags[i] & WORD_UNIQUE)) {
	wordflags[i] += WORD_UNIQUE;
	p = words[i];  n = wordprev[i];
	while(words[n]) {
	    n = wordprev[k = n];
	    if(word_cmp(k, p, -1)) continue;
	    demote_word(k);
	    str_assign(words + k, (char *)0);
	    wordflags[i] |= wordflags[k];	/* move word flags */
	    if((wordflags[k] &= WORD_UNIQUE)) break;
	}
    }
}

/*
 * compare at most len chars of i-th word and str, either case sensitive
 * or insensitive (depending on word flags). return 0 if they match.
 */
static int word_cmp(i, str, len)
int i; char *str; int len;
{
    register int c1, c2, n = len;
    register char *a, *b, diff;
    if(n) {
	a = words[i]; b = str;
	if(wordflags[i] & WORD_CASE)
	    return strncmp(a, b, n);
	do {
	    c1 = (unsigned char)*a;
	    c2 = (unsigned char)*b;
	    /* this may, or may not be lexical diff... */
	    diff = tolower(c1) - tolower(c2);
	} while(!diff && --n && (++a, ++b, c2));
	n = diff;	/* we don't care how it sign extends */
    }
    return n;
}

/*
 * reinsert a node just before wordindex ie make it the last word
 */
static void demote_word(i)
int i;
{
    int j = wordprev[i], k = wordnext[i];
    wordprev[k] = j; wordnext[j] = k;	/* remove from list */
    j = wordindex; k = wordnext[j];
    wordprev[i] = j; wordnext[i] = k;	/* prepare for insert */
    wordprev[k] = i; wordnext[j] = i;	/* and reinsert */
}

/*
 * insert word (with length n) to word completion ring
 * TODO: allow the user to add permanent entries
 */
void put_word(s, n)
char *s; int n;
{
    int r = wordindex;
    words[r] = strn_dup(s, n);  wordflags[r] = 0;
    while(wordflags[r = wordnext[r]] & WORD_RETAIN);
    demote_word(r);  wordindex = r;
    str_assign(words + r, (char *)0);
}

/*
 * add a permanent word to completion list
 * called at startup to set up the #commands
 */
void put_word_permanent(s, n)
char *s; int n;
{
    static int count;
    int i = count + 1;
    if(i < WORDSKEPT) {
	words[i] = strn_dup(s, n);
	wordflags[i] = WORD_UNIQUE | WORD_RETAIN | WORD_CASE;
	count = i;
    }
}

/*
 * initialize the word completion ring
 */
void init_completions()
{
    int j = 0, k = WORDSKEPT - 1;
    do {
	wordprev[k] = j; wordnext[j] = k;
    } while(j = k, --k >= 0);
}

