/*
 * expr.c
 * source file for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/* NOTE: use something else if expr_value typedef if not long */
long strtol();

#include "cancan.h"
#include "parse.h"
#include "ctypes.h"

static char *expr_get_token();

typedef enum {
    T_MUL, T_DIV, T_MOD, T_ADD, T_SUB, T_XOR,
    T_LT, T_GT, T_AND, T_OR,
    T_EQ, T_NE, T_SHL, T_SHR,
    T_LE, T_GE, T_LAND, T_LOR,
    T_NOP
} expr_token;

static char const op_string[] = {
    '*', '/', '%', '+', '-', '^',
    '<', '>', '&', '|',
    '=', '!', 0
};

static char const op_precedence[] = {
    10, 10, 10, 9, 9, 4,
    7, 7, 5, 3,
    6, 6, 8, 8,
    7, 7, 2, 1,
    0
};

#define N_PR	11		/* number of precedence levels */

static char *expr_get_token(str, tok)
char *str;
expr_token *tok;
{
    expr_token t;
    char *s = str;
    while(*s == ' ') s++;
    if(*s && (str = strchr(op_string, *s))) {
	t = str - op_string; s++;
	switch(t) {
	    case T_EQ:
	    case T_NE:
		if(*s == '=') s++;
		else t = T_NOP;
		break;
	    case T_LT:
		if(*s == '<') { t += T_SHL - T_LT;  s++; }
		else if(*s == '=') { t += T_LE - T_LT;  s++; }
		break;
	    case T_GT:
		if(*s == '>') { t += T_SHR - T_GT;  s++; }
		else if(*s == '=') { t += T_GE - T_GT;  s++; }
		break;
	    case T_AND:
		if(*s == '&') { t += T_LAND - T_AND;  s++; }
		break;
	    case T_OR:
		if(*s == '|') { t += T_LOR - T_OR;  s++; }
		break;
	    default:
		break;
	}
    } else
	t = T_NOP;
    *tok = t;
    return s;
}

/*
 * calculate the value of an expression. signed arithmetic only.
 * return advanced string pointer, or NULL in case of an error.
 */
char *expr_calc(str, val)
char *str;
expr_value *val;
{
    expr_value v[N_PR];
    expr_token op[N_PR];
    char *s = str;
    int n = 0;
    do {
	s = expr_getval(s, v + n);
	if(!s) break;
	s = expr_get_token(s, op + n);
	while(n && op_precedence[op[n - 1]] >= op_precedence[op[n]]) {
	    expr_value a, b;
	    n--; a = v[n]; b = v[n + 1];
	    switch(op[n]) {
		case T_EQ: a = a == b; break;
		case T_NE: a = a != b; break;
		case T_LE: a = a <= b; break;
		case T_GE: a = a >= b; break;
		case T_LT: a = a < b; break;
		case T_GT: a = a > b; break;
		case T_LOR: a = a || b; break;
		case T_LAND: a = a && b; break;
		case T_SHL: a = a << b; break;
		case T_SHR: a = a >> b; break;
		case T_OR: a = a | b; break;
		case T_AND: a = a & b; break;
		case T_XOR: a = a ^ b; break;
		case T_ADD: a = a + b; break;
		case T_SUB: a = a - b; break;
		case T_MUL: a = a * b; break;
		case T_DIV: if(!b) return 0;
			    a = a / b; break;
		case T_MOD: if(!b) return 0;
			    a = a % b; break;
		default: break;
	    }
	    v[n] = a; op[n] = op[n + 1];
	}
    } while(op[n++] != T_NOP);
    *val = v[0];
    return s;
}

/*
 * return string pointer properly advanced, NULL signifies failure.
 */
char *expr_getval(str, val)
char *str;
expr_value *val;
{
    char *u_s, *s = str;
    while(*s == ' ' || *s == '+' || *s == '-' || *s == '~' || *s == '!') s++;
    if(*(u_s = s) == '(') {
	s = expr_calc(s + 1, val);
	if(!s || *s++ != ')') return 0;
    } else {
	int n = digit_span(s);
	if(!n) {
	    n = label_span(s);
	    if(!n) return 0;
	    s = get_envn(s, n);
	    *val = 0;
	}
	if(s) *val = strtol(s, (char **)0, 10);
	s = u_s + n;
    }
    while(u_s != str)			/* handle unary operators */
	if(*--u_s == '-') *val = -*val;
	else if(*u_s == '!') *val = !*val;
	else if(*u_s == '~') *val = ~*val;
    return s;
}

/*
 * same as expr_getval, but accept a simple number only
 */
char *expr_getnum(str, val)
char *str;
expr_value *val;
{
    char *s = str;
    if(!*s) return 0;
    *val = strtol(s, &str, 10);
    if(*(s = str)) {
	if(*s != ' ') return 0;
	do s++; while(*s == ' ');
    }
    return s;
}

/*
 * stringify val and copy it to buf. return buf.
 */
char *expr_val2str(buf, val)
char *buf;
expr_value val;
{
    char sbuf[4 * sizeof(expr_value) + 4];
    char *s = sbuf + sizeof(sbuf), sign = 0;
    expr_value v = val;
    if(v < 0) { sign = '-'; v = -v; }
    *--s = '\0';
    do {
	*--s = i_to_digit(v % (unsigned)10);
	v = v / (unsigned)10;
    } while(v);
    if(sign) *--s = sign;
    return strcpy(buf, s);
}

/*
 * same as atoi, but our own wrapper
 */
int str_atoi(str)
char *str;
{
    return (int)strtol(str, (char **)0, 10);
}

