/*
 * list.c - list utility functions for cancan
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "cancan.h"
#include "list.h"

/*
 * NOTE: maybe we should keep a single hashed symbol table, like the
 * foo{thing} in perl. The way action matching and marker search work,
 * however, we still traverse those lists all the time.
 */

static void ktree_insert();
static void ktree_remove();
static unsigned int hash_val();

ktreenode *keymap[256];		/* table for keyseq lookups */
varnode *varhash[HASH_BUCKETS];
aliasnode *cmdhash[HASH_BUCKETS];

aliasnode *recording_alias;	/* node ptr to recording alias, or 0 */
aliasnode *aliases;		/* head of alias list */
keynode *keydefs;		/* head of key binding list */
actionnode *actions;		/* head of action list */
marknode *markers;		/* head of marker list */
#ifndef OMIT_SCHEDULE
schedulenode *schedule;		/* head of scheduled commands list */
#endif

/*
 * TODO: allocate nodes from separate trays.
 */

/*
 * create a new node and set its label
 */
defnode *new_node(size, key)
int size; char *key;
{
    defnode *new = (defnode *)malloc(size);
    if(!new) syserr("malloc");
    new->key = str_dup(key);
    return new;
}

/*
 * free the node created by new_node()
 */
void free_node(node)
defnode *node;
{
    free(node->key);
    free((char *)node);
}


/*
 * generic list node lookup routine
 * return pointer to pointer to node or a pointer to NULL if no match
 */
defnode **list_lookup(base, key)
defnode **base;
char *key;
{
    register defnode **p;
    register int c = key[0];
    for(p = base; *p; p = &(*p)->next)
	if((char)c == (*p)->key[0] && !strcmp(key, (*p)->key)) break;
    return p;
}

/*
 * insert a node into a list
 */
defnode **list_insert(base, node)
defnode **base, *node;
{
#ifdef SORT_BY_LABEL
    while((*base) && strcmp((*base)->key, node->key) < 0)
	base = &(*base)->next;
#endif
    node->next = *base;
    *base = node;
    return base;
}

/*
 * remove a node from a list
 * return pointer to node, so caller can free it.
 */
/*static inline*/
defnode *list_remove(base)
defnode **base;
{
    defnode *p = *base;
    *base = p->next;
    return p;
}


/*
 * generic hash lookup routine
 */
hashnode **hash_lookup(table, key, len)
hashnode **table;
char *key; int len;
{
    hashnode **base = table + hash_val(key, len) % HASH_BUCKETS;
    for( ; *base; base = &(*base)->h_next)
	if((*base)->klen == len && !memcmp((*base)->key, key, len)) break;
    return base;
}

/*
 * insert the node into a hash
 * note that hashnodes can be inserted both in a hash and a list
 */
hashnode **hash_insert(table, node)
hashnode **table, *node;
{
    hashnode **base = table + hash_val(node->key, node->klen) % HASH_BUCKETS;
    node->h_next = *base;
    *base = node;
    return base;
}

/*
 * remove the node from hash
 * return pointer to node, so caller can free it
 */
/*static inline*/
hashnode *hash_remove(base)
hashnode **base;
{
    hashnode *p = *base;
    *base = p->h_next;
    return p;
}

/*
 * return the hash value of a string
 * NOTE: we are using { x(i+1) = 3*x(i) + str[i] (mod 128) } hval, which
 * may be a poor choice, if you say so. In a (very) few tests i run, it
 * seemed reasonable, though.
 */
static unsigned int hash_val(str, len)
char *str; int len;
{
    unsigned int x = 0;
    if(len) do {
	unsigned char c = *str++;
	x = 3 * x + c;
    } while(--len);
    return x;
}


/*
 * insert a leaf node into keyseq tree, allocating
 * internal nodes as necessary.
 */
static void ktree_insert(root, key, leaf)
ktreenode **root;
char *key;
struct ktree_leaf *leaf;
{
    while(*key) {
	if(*root) {
	    if(!(*root)->cval) return;		/* a leaf node!? */
	    if((*root)->cval != *key)
		{ root = &(*root)->tree.right; continue; }
	} else {
	    struct ktree_internal *new = (struct ktree_internal *)
		    malloc(sizeof(struct ktree_internal));
	    if(!new) syserr("malloc");
	    new->cval = *key;
	    new->right = 0; new->child = 0;
	    *root = (ktreenode *)new;
	}
	root = &(*root)->tree.child;
	key++;
    }
    if(!*root) *root = (ktreenode *)leaf;
}

/*
 * remove the leaf node associated with a key.
 */
static void ktree_remove(root, key)
ktreenode **root;
char *key;
{
    for( ; *root; root = &(*root)->tree.right) {
	if((*root)->cval) {	/* an internal node */
	    if((*root)->cval != *key) continue;
	    ktree_remove(&(*root)->tree.child, key + 1);
	    if(!(*root)->tree.child) {
		ktreenode *old = *root;
		*root = old->tree.right;
		free((char *)old);
	    }
	} else if(!*key) {	/* found leaf node */
	    *root = 0;
	}
	break;
    }
}


aliasnode *new_aliasnode(key, len)
char *key; int len;
{
    aliasnode *p = (aliasnode *)new_node((int)sizeof(aliasnode), key);
    p->nlen = len; p->subst = 0; p->func = 0; p->locked = 0;
    return p;
}

/*
 * add a node to the alias list,
 * also inserting it into the cmdhash
 */
aliasnode **add_aliasnode(name, subst)
char *name, *subst;
{
    int len = str_len(name);
    aliasnode *p = *lookup_cmdhash(name, len);
    if(!p) {
	p = new_aliasnode(name, len);
	hash_insert_(cmdhash, p);
    }
    str_assign(&p->subst, subst);
    return (aliasnode **)list_insert_(&aliases, p);
}

void delete_aliasnode(base)
aliasnode **base;
{
    aliasnode *p = (aliasnode *)list_remove_(base);
    str_assign(&p->subst, (char *)0);
    if(!p->func && !p->locked)
	free_node_(hash_remove_(lookup_cmdhash(p->name, p->nlen)));
}


varnode *new_varnode(key, len)
char *key; int len;
{
    varnode *p = (varnode *)new_node((int)sizeof(varnode), key);
    p->nlen = len;
    return p;
}

varnode *make_varlist()
{
    varnode *p, *base = 0;
    int i;
    for(i = 0; i < HASH_BUCKETS; i++)
	for(p = varhash[i]; p; p = p->h_next)
	    list_insert_(&base, p);
    return base;
}


/*
 * look up an action node by pattern
 */
actionnode **lookup_action_pattern(pattern)
char *pattern;
{
    actionnode **p = &actions;
    while(*p && strcmp(pattern, (*p)->pattern))
	p = &(*p)->next;
    return p;
}

/*
 * add a node to the action list
 */
actionnode **add_actionnode(name, pattern, commands)
char *name, *pattern, *commands;
{
    actionnode *new = (actionnode *)new_node((int)sizeof(actionnode), name);
    new->pattern = str_dup(pattern);
    new->commands = str_dup(commands);
    new->active = 0;
    return (actionnode **)list_insert_(&actions, new);
}

void delete_actionnode(base)
actionnode **base;
{
    actionnode *p = (actionnode *)list_remove_(base);
    free(p->commands);
    free(p->pattern);
    free_node_(p);
}


/*
 * add a node to the marker list
 */
marknode **add_marknode(pattern, attrcode)
char *pattern; int attrcode;
{
    marknode *new = (marknode *)new_node((int)sizeof(marknode), pattern);
    new->plen = str_len(new->pattern);
    new->attrcode = attrcode;
    return (marknode **)list_insert_(&markers, new);
}

void delete_marknode(base)
marknode **base;
{
    free_node(list_remove_(base));
}


/*
 * add a node to the keybindings list,
 * also inserting keydef in the keymap
 */
keynode **add_keynode(name, seq, value)
char *name, *seq, *value;
{
    keynode *new = (keynode *)new_node((int)sizeof(keynode), name);
    new->sequence = str_dup(seq);
    new->type = 0;
    new->keydef.cval = 0;
    new->keydef.str = str_dup(value);
    new->keydef.func = 0;
    ktree_insert(keymap + (unsigned char)*seq, seq + 1, &new->keydef);
    keymap[0] = keymap[0200]; /* avoid test for 0 in line_edit */
    return (keynode **)list_insert_(&keydefs, new);
}

void delete_keynode(base)
keynode **base;
{
    keynode *p = (keynode *)list_remove_(base);
    char *seq = p->sequence;
    ktree_remove(keymap + (unsigned char)*seq, seq + 1);
    keymap[0] = keymap[0200]; /* avoid test for 0 in line_edit */
    free(p->keydef.str);
    free(p->sequence);
    free_node_(p);
}


#ifndef OMIT_SCHEDULE
/*
 * add a node to the schedule list (sorted by delay)
 */
schedulenode **add_schedulenode(label, commands, delay)
char *label, *commands; long delay;
{
    schedulenode *new =
	    (schedulenode *)new_node((int)sizeof(schedulenode), label);
    new->commands = str_dup(commands);
    new->delta = 0;
    new->next = schedule; schedule = new;
    return reschedule(&schedule, delay);
}

/*
 * reinsert a schedulenode with new delay, keeping the list sorted
 * minus sign of delay indicates that the current delay is increased
 * rather than set to a new value
 */
schedulenode **reschedule(np, delay)
schedulenode **np;  long delay;
{
    schedulenode *old = *np;
    if((*np = old->next))
	(*np)->delta += old->delta;
    if(delay > 0)
	np = &schedule;
    else
	delay = (old->delta > 0 ? old->delta : 0) - delay;
    /* if(delay < 0) ; else */
    for( ; *np && (*np)->delta <= delay; np = &(*np)->next)
	delay -= (*np)->delta;
    old->delta = delay;
    if((old->next = *np))
	old->next->delta -= delay;
    *np = old;
    return np;
}

void delete_schedulenode(base)
schedulenode **base;
{
    schedulenode *p = (schedulenode *)list_remove_(base);
    if(*base) (*base)->delta += p->delta;
    free(p->commands);
    free_node_(p);
}
#endif

