/*
 * list.h
 * linked list operations and routines for other data structures.
 */

#define HASH_BUCKETS 128

/* generic linked list node */
typedef struct defnode {
    struct defnode *next;
    char *key;
} defnode;

/* generic hash node (built on defnode) */
typedef struct hashnode {
    struct hashnode *next;	/* next in some list */
    char *key; int klen;
    struct hashnode *h_next;	/* next in bucket */
} hashnode;


typedef struct cmdhash_node {
    struct cmdhash_node *next;
    char *name; int nlen;
    struct cmdhash_node *h_next;
    char *subst;
    cmd_func *func;
    int locked;
} aliasnode;


/*
 * NOTE: varnode.name contains the whole "name=value" pair, with
 * varnode.nlen being the length of the name part. Only the hash
 * is kept up to date as vars are modified; the (possibly sorted)
 * list of variables is recreated each time prior to use.
 */
/* typedef char* (var_func)(); */
typedef struct varnode {
    struct varnode *next;
    char *name; int nlen;
    struct varnode *h_next;
    /* var_func *func; */
} varnode;


/* keybind tree internal node */
struct ktree_internal {
    char cval;			/* must be !0 */
    union ktree_node *right;	/* same level node */
    union ktree_node *child;	/* child node */
};

/* keybind tree leaf node ie {seq ==> keydef} */
struct ktree_leaf {
    char cval;			/* must be 0 */
    char *str;			/* passed along to func */
    editfunc *func;		/* edit function */
};

typedef union ktree_node {
    char cval;
    struct ktree_leaf leaf;
    struct ktree_internal tree;
} ktreenode;

enum key_type { KEY_NORMAL = 0, KEY_DEFAULT = 1, KEY_PROTECTED = 2 };

typedef struct keynode {
    struct keynode *next;
    char *label;
    char *sequence;
    int type;
    struct ktree_leaf keydef;
} keynode;


/*
 * NOTE: a negative actionnode.active means action is disabled.
 * Otherwise, its value is the minimum length of matchable text.
 */
typedef struct actionnode {
    struct actionnode *next;
    char *label;
    char *pattern;
    char *commands;
    int active, mask;
} actionnode;


typedef struct marknode {
    struct marknode *next;
    char *pattern; int plen;
    int attrcode;
} marknode;


typedef struct schedulenode {
    struct schedulenode *next;
    char *label;
    char *commands;
    long delta;
} schedulenode;


defnode *new_node();
void free_node();

defnode **list_lookup();
defnode **list_insert();
defnode *list_remove();

hashnode **hash_lookup();
hashnode **hash_insert();
hashnode *hash_remove();

#define free_node_(node)	free_node((defnode *)node)
#define list_lookup_(b, key)	list_lookup((defnode **)b, key)
#define list_insert_(b, node)	list_insert((defnode **)b, (defnode *)node)
#define list_remove_(b)		list_remove((defnode **)b)
#define hash_lookup_(t, key, len) hash_lookup((hashnode **)t, key, len)
#define hash_insert_(t, node)	hash_insert((hashnode **)t, (hashnode *)node)
#define hash_remove_(b)		hash_remove((hashnode **)b)

aliasnode *new_aliasnode();
#define lookup_cmdhash(key, len) ((aliasnode **)hash_lookup_(cmdhash, key, len))
#define lookup_alias(name) ((aliasnode **)list_lookup_(&aliases, name))
aliasnode **add_aliasnode();
void delete_aliasnode();

#define lock_aliasnode(np) (np)->locked++
#define release_aliasnode(np) \
    do {				\
	aliasnode *tp = (np);		\
	tp->locked--;			\
	if(tp->subst) break;		\
	tp->next = tp;			\
	delete_aliasnode(&tp->next);	\
    } while(0)

varnode *new_varnode();
varnode *make_varlist();
#define lookup_varhash(key, len) ((varnode **)hash_lookup_(varhash, key, len))

#define lookup_key(name) ((keynode **)list_lookup_(&keydefs, name))
keynode **add_keynode();
void delete_keynode();

#define lookup_action(name) ((actionnode **)list_lookup_(&actions, name))
actionnode **lookup_action_pattern();
actionnode **add_actionnode();
void delete_actionnode();

#define lookup_marker(name) ((marknode **)list_lookup_(&markers, name))
marknode **add_marknode();
void delete_marknode();

#define lookup_schedule(name) ((schedulenode **)list_lookup_(&schedule, name))
schedulenode **reschedule();
schedulenode **add_schedulenode();
void delete_schedulenode();

extern ktreenode *keymap[256];
extern varnode *varhash[HASH_BUCKETS];
extern aliasnode *cmdhash[HASH_BUCKETS];
extern aliasnode *recording_alias;
extern aliasnode *aliases;
extern keynode *keydefs;
extern actionnode *actions;
extern marknode *markers;
extern schedulenode *schedule;
