/*
 * parse.c
 * command parsing, argument parsing routines for cancan
 */

#include <stdio.h>
#include <string.h>

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "ctypes.h"

static int subst_param();
static void record_line();
static struct param *match_action();
static int exec_action();
static void split_args();
static int arg_span();
static char *skip_arg();

/*
 * silent = 1 if the line should not be echoed, 0 otherwise.
 * if invoked by alias or action, params are passed along.
 */
void parse(cmdline, silent, param)
char cmdline[LINELEN];
int silent;
struct param param[10];
{
    static int alias_depth;
    char cmdbuf[LINELEN], argbuf[LINELEN];
    struct parse_context ctx;
    char *cmd;
    char *verb; int vlen;
    char *args; int alen;
    aliasnode *np;
    enum { is_text, is_alias, is_command } what;

    if(!brace_balance(cmdline)) return;

    ctx.silent = silent;
    ctx.subst = 0;
    ctx.params = param;

    cmd = skipspace(cmdline);
    if(*cmd == '{') {				/* line enclosed in braces? */
	char *s = skip_arg(++cmd, '}');
	if(!*skipspace(s + 1)) {
	    memcpy(cmdbuf, cmd, s - cmd);
	    cmdbuf[s - cmd] = 0;
	    ctx.subst = 1;
	}
    }
    if(!ctx.subst) strcpy(cmdbuf, cmdline);
    cmd = cmdbuf;

    /* PROCESS THE COMMANDS ONE BY ONE */
    do {
	ctx.cmdchain = chop_arg(cmd, CMDSEP);	/* chop into commands */

	verb = skipspace(cmd);			/* find the first word */
	vlen = arg_span(verb);
	/* skip some processing if !vlen? */
	np = *lookup_cmdhash(verb, vlen);	/* and what do we have? */
	args = verb + vlen;
	if(np && np->subst && !np->locked) {
	    what = is_alias;
#if 0	/* all commands start with a '#' anyway */
	} else if(np && np->func) {
	    what = is_command;
#endif
	} else if(*verb == '#') {
	    what = is_command;
	} else {
	    args = cmd;
	    what = is_text;
	}

	if(what != is_command && recording_alias && !ctx.params)
	    record_line(cmd);

	if(!ctx.subst) {
	    alen = str_copy(argbuf, args);
	    if(alias_depth && !ctx.cmdchain)	/* append arguments */
		if(ctx.params && ctx.params[0].len) {
		    int n = sizeof(argbuf) - alen;
		    if(n > ctx.params[0].len)
			n = ctx.params[0].len;
		    memcpy(argbuf + alen, ctx.params[0].s, n);
		    alen += n;
		}
	} else
	    alen = subst_param(argbuf, args, ctx.params);

	if(alen > (int)sizeof(argbuf) - 1) {
	    alen = sizeof(argbuf) - 1;
	    printf("#warning: line truncated\n");
	}
	args = argbuf; args[alen] = 0;

	if(what == is_text) {			/* simple text */
	    if(alen) {
		if(alen == 1 || OPTION(OPT_SPEEDWALK))
		    if(parse_walk(args)) continue;
		if(!ctx.subst) alen = unescape(args);
	    }
	    if(!ctx.silent)
		printf("[%s%s%s]\n", edattrbeg, args, edattrend);
	    send_line(args, alen);
	    continue;
	}
	if(what == is_alias) {			/* process an alias */
	    struct param new_param[10];
	    if(alias_depth < LOOP_LIMIT) {
		++alias_depth;
		lock_aliasnode(np);		/* block recursion */
		split_args(args, alen, new_param);
		parse(np->subst, ctx.silent, new_param);
		release_aliasnode(np);
		--alias_depth;
	    }
	    continue;
	}

	while(*args == ' ')			/* find args start */
	    { args++; alen--; }
	if(np && np->func) {			/* cancan #command */
	    (*np->func)(args, &ctx);
	} else if(vlen <= 1) {			/* a comment: # */
	    continue;
	} else if(digit_span(verb + 1) == vlen - 1) {	/* #17 ... */
	    int i = str_atoi(verb + 1);
	    if(i > LOOP_LIMIT) i = LOOP_LIMIT;
	    if(i > 0) do {
		parse(args, ctx.silent, (struct param *)0);
	    } while(--i);
	    /* else printf("#bogus repeat count\n"); */
	} else {
	    verb[vlen] = 0;
#ifndef OMIT_SESSIONS
	    if(verb[1] == '#') {		/* ##session ... */
		if(!ctx.subst) alen = unescape(args);
		args[alen] = '\n';
		send_to_sess(verb + 2, args, alen + 1);
		continue;
	    }
#endif
	    printf("#unknown cancan command '%s'\n", verb);
	}
    } while((cmd = ctx.cmdchain));
}

/*
 * execute walk if word is valid [speed]walk sequence -
 * return 1 if walked, 0 if not
 */
int parse_walk(word)
char *word;
{
    static char walk_dir;
    char c, *s = word;
    int llim;

    do { c = *s; } while(iswalkseq(c) && *++s);
    if(!iswalkdir(c)) return 0;

    llim = LOOP_LIMIT; s = word;
    do {
	int n = 1;
	if(isdigit(*s))
	    { n = str_atoi(s); s += digit_span(s); }
	for(walk_dir = *s; n; n--) {
	    send_line(&walk_dir, 1);
	    add_dir_to_map(*s);
	    if(!--llim) break;
	}
    } while(llim && *++s);
    return 1;
}

/*
 * if alias recording is active, record given line in that alias
 */
static void record_line(line)
char *line;
{
    static int i;
    aliasnode *np = recording_alias;	/* assert(np && np->locked); */
    if(!np->subst) {
	i = 0;
	add_aliasnode(np->name, line);
    } else {
	/*
	 * NOTE: lines that exceed LINELEN (with "#alias foo=" prefix),
	 * are not recorded.
	 */
	char buf[LINELEN], *new_line = buf;
	int len = str_copy(new_line, np->subst);
	int room = LINELEN - 10 - str_len(line);
	if(len >= room - LINELEN / 2) {	/* leave ample room for editing */
	    do {
		sprintf(new_line, "%s%d", np->name, ++i);
	    } while(*lookup_alias(new_line));
	    len = str_len(new_line);
	    if(len >= room) return;
	    add_aliasnode(new_line, np->subst);
	}
	new_line[len] = ';';
	strcpy(new_line + len + 1, line);
	str_assign(&np->subst, new_line);
    }
}

/*
 * verify that the braces in a line are properly balanced
 */
int brace_balance(str)
char *str;
{
    char *message = "#unbalanced %c near \"%.16s\"\n";
    char *d, *s = str;
    if(skip_arg(s, '\0')) return 1;
    if((d = skip_arg(s, '}')))
	printf(message, *d, (d - str > 15) ? d - 15 : str);
    else {
	do {
	    d = skip_arg(s, '{');
	    s = skip_arg(d + 1, '}');
	} while(s++);
	printf(message, *d, d);
    }
    return 0;
}

/*
 * {} *must* be balanced in input string. input will be trashed.
 * leave {...} contents intact, otherwise unescape and replace
 * &0 .. &9, &label, &{label}, &[expression], &(command).
 */
static int subst_param(buf, str, param)
char *buf, *str;
struct param param[10];
{
    char sbuf[LINELEN];
    char *src, *s = str;
    int i, room = LINELEN;

    while(*s) {
	if(*s != '&') {
	    if(*s != '{') {
		if(*s == '\\' && !*++s) break;
		if(room) --room, *buf++ = *s;
		s++; continue;
	    } else {
		src = s;  s = skip_arg(s + 1, '}') + 1;
		i = s - src;
	    }
	} else switch(*++s) {
	case '{':				/* &{variable}	*/
	    s = skip_arg((src = s + 1), '}');
	    *s++ = 0;
	    i = subst_param(sbuf, src, param);
	    if(i >= LINELEN) continue;
	    src = get_envn(sbuf, i);
	    if(!src) continue;
	    i = -1; break;
	case '[':				/* &[expr]	*/
	    s = skip_arg((src = s + 1), ']');
	    if(s) {
		expr_value val;
		*s++ = 0;
		i = subst_param(sbuf, src, param);
		if(i >= LINELEN) continue;
		sbuf[i] = 0;
		src = expr_calc(sbuf, &val);
		if(!src || *src) continue;
		src = expr_val2str(sbuf, val);
		i = -1; break;
	    }
	    s = src - 1;
	    printf("#unbalanced %c near \"%.16s\"\n", *s, s - 1);
	    if(room) --room, *buf++ = '&';
	    continue;
	case '(':				/* &(command)	*/
	    s = skip_arg((src = s + 1), ')');
	    if(s) {
		FILE *fp;
		*s++ = 0;
		unescape(src);
		if((fp = do_popen(src, "r"))) {
		    if((i = fread(buf, 1, room, fp)) > 0) {
			if(buf[i - 1] == '\n') i--;
			buf += i;  room -= i;
		    }
		    do_pclose(fp);
		}
		continue;
	    }
	    s = src - 1;
	    printf("#unbalanced %c near \"%.16s\"\n", *s, s - 1);
	    if(room) --room, *buf++ = '&';
	    continue;
	default:
	    if((unsigned)(i = digit_value(*s)) < 10) {
		s++;
		if(!param || !param[i].len) continue;
		src = param[i].s;
		i = param[i].len;
		break;
	    }
	    if((i = label_span(s))) {
		src = s; s += i;
		src = get_envn(src, i);
		if(!src) continue;
		i = -1; break;
	    }
	    if(room) --room, *buf++ = '&';
	    continue;
	} /* switch ends */
	if(i < 0) i = str_len(src);
	if(room < i) i = room;
	memcpy(buf, src, i);
	buf += i;  room -= i;
    } /* end while */
    return LINELEN - room;
}


/*
 * scan through the pattern, determine what parameters it sets
 * and what is the length of a shortest string that can match.
 * NOTE: maybe we should go on, and compile the pattern here..
 */
void study_action_pattern(p)
actionnode *p;
{
    int prm, mask = 0x3ff, n = 0;
    char *s = p->pattern;
    if(*s == '^') s++;
    for( ; *s; s++)
	if(*s != '&') n++;
	else {
	    while(*++s == '&');
	    prm = digit_value(*s);
	    if((unsigned)prm >= 10) prm = 0, s--;
	    mask &= ~(1 << prm);
	}
    if(p->active < 0) n = -1 -n;
    p->active = n;
    p->mask = mask;
}

/*
 * match action containing &0..&9 and return actual params in param array
 *
 * NOTE: if pattern contains a sequence of &n-s, all but last are ignored.
 * FEATURE: &&n will match a single word and put the result in &n.
 * NOTE: '&' shouldn't match an empty word. Only printable chars should be
 * allowed to match a & or &n. Should try to find longest possible match.
 * FEATURE: can't match '&' as plain text in pattern (its always special).
 * FEATURE: no backtracking used, therefore a match with & might fail, even
 * though respective regexp would not, e.g. pat:"a & b", src:"a a a b" fails.
 * TODO: wildcard match: count '&'s like regex groups are counted, drop word
 * match, add end of line anchor '$'. Recognize escaping '\'. Compile it.
 * TODO: regex match: "#action /ERE/={foo}". Match action against whole line
 * if $-anchored, and against fragments otherwise? Support attributes: /ERE/i
 * case insensitive, /ERE/l to force match against full lines? Etc.
 * TODO: compile all actions and markers into one mighty DFA.
 */
static struct param *match_action(pat, src)
char *pat, *src;
{
    static struct param param[10];
    char *mpat = insertbuf;	/* use insertbuf for tmp buf space. hm */
    register char *s = src, *p = pat;
    int k, prm;
    char mflags = 1;	/* 1 - at pattern beginning, 2 - single word match */
    if(*p == '^') {			/* anchor match at line start */
	if(col0) return 0;
	for(p++; *p && *p != '&'; s++, p++)
	    if(*s != *p) return 0;
	mflags = 0;
    }
    prm = 0;
    while(*p) {
	if(*p == '&') do {		/* pick out parameters */
	    p++;  prm = digit_value(*p++);
	    if((unsigned)prm >= 10)	/* support single & for back compat */
		p--, prm = 0, mflags |= 2;	/* match a single word only */
	} while(*p == '&');
	if(*p) {			/* attempt match */
	    register char *mp;
	    param[prm].s = s;
	    s = strchr(s, *p);
	    if(!s) return 0;
	    mp = mpat;
	    do { *mp++ = *p++; } while(*p && *p != '&');
	    *mp = 0;
	    s = strstr(s, mpat);
	    if(!s) return 0;
	    param[prm].len = s - param[prm].s;
	    s += mp - mpat;

	    if(mflags & 2) {		/* post-process single word match */
		char *w;
		k = param[prm].len;  w = param[prm].s;
		while(k && isword(w[k - 1])) k--;
		if(k) {
		    if(!(mflags & 1)) return 0;
		    /* unanchored '&' start, take last word */
		    param[prm].len -= k;  param[prm].s += k;
		}
	    }
	    mflags = 0;
	    continue;
	}
	/* '&' at the end of the pattern space */
	if(mflags & 2) {	/* take first word */
	    if(mflags & 1) while(*s && !isword(*s)) s++;
	    for(k = 0; isword(s[k]); k++);
	} else
	    k = str_len(s);
	param[prm].len = k;  param[prm].s = s;
	break;
    } /* end while */
    return param;
}

/*
 * execute an action. return 1 if line should be printed,
 * or 0 if it should be suppressed.
 */
static int exec_action(p, param)
actionnode *p;
struct param *param;
{
    unsigned int mask = p->mask;
    struct param *pp = param;
    if(!strcmp("#gag", p->commands))
	return 0;
    while((mask >>= 1)) /* clear the params we did not set */
	if(pp++, mask & 1) pp->len = 0;
    clear_prompt();
    parse(p->commands, 0, param);
    check_line0();
    redraw_prompt();
    return 1;
}

/*
 * search for actions to trig on an input line.
 * NOTE: a negative p->active means action is disabled. Otherwise,
 * its value is the minimum length of matchable text.
 */
int search_action(line, len)
char *line;
int len;
{
    register unsigned int l = len;
    register actionnode *p;
    for(p = actions; p; p = p->next)
	if((unsigned)p->active <= l) {
	    struct param *match = match_action(p->pattern, line);
	    if(!match) continue;
	    /* set &0 to whole line if it wasn't used in pattern */
	    if(p->mask & 1) { match[0].len = l; match[0].s = line; }
	    return exec_action(p, match);
	}
    return 1;
}


/*
 * enclose in braces and escape &\; or unbalanced braces, when needed.
 * otherwise just copy the string.
 * NOTE: perverse definitions that need to be doubly escaped should be
 * disallowed in the first place (but consider #record).
 */
int escape_command(dst, src)
char *dst, *src;
{
    int room;
    char *s = src, *d = dst;
    if(!skip_arg(s, CMDSEP) && skip_arg(s, '\0'))
	return str_copy(d, s);
    *d++ = '{';
    for(room = LINELEN - 1; --room && *s; *d++ = *s++)
	if(strchr("\\{};&", *s)) {
	    if(!--room) break;
	    *d++ = '\\';
	}
    if(room) *d++ = '}';
    *d = '\0';
    return d - dst;
}

/*
 * escape special chars with a backslash, return length of result
 * NOTE: we don't check for overflow here, it shouldn't happen
 */
int escape_specials(dst, src, spec)
char *dst, *src, *spec;
{
    char *s = src, *d = dst;
    for( ; *s; *d++ = *s++)
	if(strchr(spec, *s)) *d++ = '\\';
    *d = 0;
    return d - dst;
}

/*
 * remove escapes (backslashes) from a string
 * return the length of result
 */
int unescape(str)
char *str;
{
    register char *s = str, *d = str, c;
    do {
	c = *s++;
	if(c == '\\') c = *s++;
    } while((*d++ = c));
    return d - str - 1;
}

/*
 * make an escape sequence printable (for use in definition-files)
 */
void escape_seq(dst, seq)
char *dst, *seq;
{
    char *s = seq, *d = dst;
    unsigned char c;
    /*
     * rules: control chars are written as ^X, other unprintables
     * in octal \nnn form. Don't bother to escape =\^ with a \,
     * just octalize 'em.
     */
    while((c = *s++)) {
	if(c >= '\177' || c == '=' || c == '\\' || c == '^') {
	    sprintf(d, "\\%03o", (int)c);
	    d += 4; continue;
	} else if(c < ' ')
	    *d++ = '^', c |= 0100;
	*d++ = c;
    }
    *d = '\0';
}

/*
 * unescape, but also recognize ^c, \nnn syntax.
 * return the length of result
 */
int unescape_seq(str)
char *str;
{
    char *s = str, *d = str, c;
    unsigned int t;
    do switch((c = *s++)) {
	case '^':
	    c = *s++;
	    if(c) {
		c = c & 037;
		if(!c) c = '\200';
	    }
	    break;
	case '\\':
	    c = *s++;
	    if((t = digit_value(c)) < 010) {
		c = t;
		if((t = digit_value(*s)) < 010) { c = c * 010; c += t; s++; }
		if((t = digit_value(*s)) < 010) { c = c * 010; c += t; s++; }
		if(!c) c = '\200';
	    }
    } while((*d++ = c));
    return d - str - 1;
}

/*
 * split str into arguments separated by whitespace.
 * parameter 0 contains the whole line.
 */
static void split_args(str, len, param)
char *str; int len;
struct param param[10];
{
    char *s = str;
    int l = len, i = 0;
    do {
	param[i].len = l;
	param[i].s = str;
	while(*s == ' ') s++;
	l = arg_span(s);
	str = s; s += l;
    } while(++i < 10);
}

/*
 * return the length of space-delimited argument.
 */
static int arg_span(str)
char *str;
{
    register char *s = str, c;
    do {
	c = *s++;
    } while(c && c != ' ');
    return s - str - 1;
}

int digit_span(str)
char *str;
{
    register int len = -1;
    do ++len; while(isdigit(str[len]));
    return len;
}

int label_span(str)
char *str;
{
    register int len = -1;
    do ++len; while(islabel(str[len]));
    return len;
}

int print_span(str)
char *str;
{
    register int len = -1;
    do ++len; while(isprint(str[len]));
    return len;
}

int get_token(args, table)
char **args, *table[];
{
    int i;
    char *word = *args;
    *args = chop_str(word, ' ');
    for(i = 0; table[i]; i++)
	if(!strcmp(word, table[i])) break;
    return i;
}

/*
 * Checks: a label may not contain a space or a '=' (escaped or not).
 * Actually, we shouldn't allow alias names like "{foo}" either.
 */
char *chop_label(str)
char *str;
{
    int escaped = 0, nested = 0;
    char *s;
    for(s = str; *s && *s != ' ' && *s != '='; s++) {
	if(escaped) { escaped = 0; continue; }
	if(*s == '\\') escaped = 1;
	if(*s == '{') nested++;
	if(*s == '}') nested--;
    }
    if(escaped || nested)	/* invalid label */
	return str;
    if(*s == ' ')		/* fix spaces after label */
	for(*s++ = 0; *s == ' '; s++);
    return s;
}

/*
 * like strchr, but honour (possibly nested) braces and escaping backslashes.
 * return pointer to the delim in str, or NULL.
 */
static char *skip_arg(str, delim)
char *str;
unsigned char delim;
{
    register int c, nested = 0;
    register char *s = str;
    do {
	do {
	    c = (unsigned char)*s++;
#if 0	    /* delim can't be islabel */
	    if(c != delim) break;
#endif
	} while(islabel(c));
	if(c == delim && !nested) break;
	if(c == '{')  { nested++; }
	if(c == '}')  { if(--nested < 0) break; }
	if(c == '\\') { c = (unsigned char)*s++; }
    } while(c);
    return (c == delim && !nested) ? (s - 1) : 0;
}

/*
 * split the argument into left and right sides, separated by delim
 * return pointer to right hand side, or 0 if no delim was found.
 */
char *chop_arg(str, delim)
char *str, delim;
{
    char *s = skip_arg(str, delim);
    if(s) *s++ = 0;
    return s;
}

/*
 * like chop_arg, but no special handling of {\;
 * used as a replacement of strtok
 */
char *chop_str(str, delim)
char *str, delim;
{
    char *s = str;
    if(*s) while(*s != delim && *++s);
    if(*s) for(*s = 0; *++s == delim; );
    return *s ? s : 0;
}

/*
 * return pointer to next non-blank char
 * NOTE: only space is considered whitespace.
 * editline is all printable chars anyway.
 */
char *skipspace(p)
char *p;
{
    while(*p == ' ') p++;
    return p;
}

