/*
 * tel.c - telnet protocol communication module for cancan
 */

#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/telnet.h>
#include <arpa/inet.h>
#include <unistd.h>

#ifdef TERM
#include <client.h>
#endif /* TERM */

#ifndef TELOPT_NAWS
#define TELOPT_NAWS	31
#endif
#ifndef TELQUAL_IS
#define TELQUAL_IS	0
#endif
#ifndef TELQUAL_SEND
#define TELQUAL_SEND	1
#endif

#ifdef TELOPTS
#define TELOPTSTR(n) ((n) > NTELOPTS ? "unknown" : telopts[n])
#endif

int socket();
int connect();

#include "cancan.h"
#include "term.h"

static void tn_sendopt();
static void tn_suboption();

#define send_raw(buf, len)	send_to_host(buf, len, -1)
#define send_binary(buf, len)	send_to_host(buf, len, -2)

enum telnetstate {
    TS_DATA = 0, TS_CR,
    TS_IAC, TS_WILL, TS_WONT, TS_DO, TS_DONT, TS_SB, TS_SBIAC
};

struct telnet_data {
    int state;
    lbuf *subopt;
    char will_binary;		/* unused */
    char will_sga;		/* unused */
    char will_naws;
    char will_ttype;
    char remote_binary;		/* unused */
    char remote_sga;
    char remote_echo;
};

#define MAXSUBOPT	20	/* currently, this is more than sufficient */

#ifndef OMIT_SESSIONS
static struct telnet_data sock_state[MAXFDSCAN];
#define SOCK_STATE(fd) sock_state[fd]
static int output_socket;	/* to which socket buffer should be sent */
#else
static struct telnet_data sock_state;
#define SOCK_STATE(fd) sock_state
#define output_socket (sockfd)
#endif

static int output_len;		/* number of characters in output_buffer */
static char output_buffer[BUFSIZE + 4];

/*
 * connect to remote host
 * NOTE: should do getservbyname() when port is not numeric.
 * NOTE: We block while connection is attempted. Keep it simple?
 */
int do_connect(addr, port)
char *addr, *port;
{
    int fd;
    set_spec_keys(1);	/* we might hang here for a long time... */

#ifndef TERM
    do {
	struct sockaddr_in address;
	int opt;
	printf("Trying %s... ", addr); fflush(stdout);
	address.sin_family = AF_INET;
#if 0	/* inet_aton() is not quite portable */
	if(!inet_aton(addr, &address.sin_addr))
#else
	/*
	 * NOTE: inet_addr is documented to return -1 for malformed requests,
	 * but it is declared to return unsigned long!  Anyway, this works.
	 */
	address.sin_addr.s_addr = inet_addr(addr);
	if((int)address.sin_addr.s_addr == -1)
#endif
	{
	    struct hostent *hp = gethostbyname(addr);
	    if(!hp) {
		printf("unknown host\n");
		fd = -1;
		break;
	    }
	    memcpy((char*)&address.sin_addr, hp->h_addr, hp->h_length);
	    address.sin_family = hp->h_addrtype;
	}

	address.sin_port = htons(str_atoi(port));
#if 0
	if(!address.sin_port) {
	    struct servent *sp = getservbyname(port, "tcp");
	    if(!sp) {
		printf("bad port\n");
		fd = -1;
		break;
	    }
	    address.sin_port = sp->s_port;
	}
#endif

	if((fd = socket(address.sin_family, SOCK_STREAM, 0)) < 0) {
	    syswarn("error creating socket");
	    break;
	}
	if(fd >= MAXFDSCAN) {
	    printf("not connected (Too many open connections)\n");
	    close(fd);  fd = -1;
	    break;
	}

#if defined(SO_OOBINLINE)
	opt = 1;
	setsockopt(fd, SOL_SOCKET, SO_OOBINLINE, (char *)&opt, sizeof(opt));
#endif
#if defined(IP_TOS)
	opt = 0x10;			/* IPTOS_LOWDELAY */
	setsockopt(fd, IPPROTO_IP, IP_TOS, (char *)&opt, sizeof(opt));
#endif
#if defined(TCP_NODELAY) && 0
	opt = 1;
	setsockopt(fd, IPPROTO_TCP, TCP_NODELAY, (char *)&opt, sizeof(opt));
#endif

	for(;;) {
	    if(connect(fd, (struct sockaddr *)&address, sizeof(address)) < 0) {
		if(errno == EINTR) continue;
		syswarn("host not responding");
		close(fd);  fd = -1;
	    } else
		printf("connected!\n");
	    break;
	}
    } while(0);

#else /* term */

    if((fd = connect_server(0)) < 0)
	printf("Unable to connect to term server\n");
    else {
	send_command(fd, C_PORT, 0, "%s:%d", addr, str_atoi(port));
	printf("Connected to term server...\n");
#ifdef TERM_COMPRESS
	send_command(fd, C_COMPRESS, 1, "y");
#endif
	send_command(fd, C_DUMB, 1, 0);
    }
#endif /* term */

    set_spec_keys(0);
    return fd;
}

void do_disconnect(fd)
int fd;
{
    struct telnet_data *tn = &SOCK_STATE(fd);
    shutdown(fd, 2);	/* this can't hurt */
    close(fd);
    tn->state = TS_DATA;
    lbuf_free(&tn->subopt);
    tn->remote_sga = tn->remote_echo = 0;
    tn->will_ttype = tn->will_naws = 0;
}

/*
 * scan through the buffer, interpret telnet protocol escapes.
 * data bytes are written back to the buffer. return their count.
 *
 * TODO: line-by-line processing. return after linemode change.
 * TODO: an "ansifilter" option would be nice as well to eliminate
 * ansi sequences/unprintables before any processing.
 */
int tn_process(buf, len)
char *buf; int len;
{
    struct telnet_data *tn = &SOCK_STATE(sockfd);
    unsigned char *s, *d, c;
    int n;

    d = s = (unsigned char *)buf;
    if((n = len)) do {
	switch(tn->state) {
	case TS_CR:
	    tn->state = TS_DATA;
	    if(*s == '\0') break;
	    /* fallthrough */

	default:
	    do {
		c = *s++;
		*d++ = c;
	    } while(--n && c != '\r' && c != IAC);

	    if(c == '\r' /* && !tn->remote_binary */) {
		tn->state = TS_CR; continue;
	    }
	    if(c == IAC) {
		d--;
		tn->state = TS_IAC; continue;
	    }
	    continue;

	case TS_IAC:
	    if(*s == IAC) {
		*d++ = IAC;
		tn->state = TS_DATA;
	    } else switch(*s) {
	    case WILL:
		tn->state = TS_WILL + (*s - WILL); break;
	    case WONT:
		tn->state = TS_WONT + (*s - WONT); break;
	    case DO:
		tn->state = TS_DO + (*s - DO); break;
	    case DONT:
		tn->state = TS_DONT + (*s - DONT); break;
	    case SB:
		tn->subopt = lbuf_alloc(MAXSUBOPT);
		tn->state = TS_SB; break;
		/*
	    case IP:
		 * NOTE: since MPI is plain text, it should be assisted
		 * by a lower level protocol escape.
	    case GA:
		 * NOTE: GA could be handled as end-of-prompt marker one day.
		 * However, prompts have to be identified as such in the
		 * absence of GA anyway.
		 */
	    default:
		/* ignore the rest of the telnet commands */
#ifdef TELOPTS
		fprintf(stderr, "[skipped IAC <%d>]\n", *s);
#endif
		tn->state = TS_DATA; break;
	    }
	    break;

	/*
	 * NOTE: perhaps we should handle some more telnet options, like:
	 * - TELOPT_BINARY: technically, this should be enabled when
	 * extended charset is used, or MP tranfer initiated.
	 * - TELOPT_LINEMODE: the proper way to switch linemode. However,
	 * it seems unnecessarily complex for our purposes, and the klugly
	 * SGA option probably needs to be processed anyway.
	 * - TELOPT_RCP: rarely used, why bother?
	 */

	case TS_WILL:
#ifdef TELOPTS
	    fprintf(stderr, "[got WILL %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_ECHO || *s == TELOPT_SGA) {
		char *topt = (*s == TELOPT_ECHO)
			? &tn->remote_echo : &tn->remote_sga;
		tn_sendopt(DO, *s, topt);
		if(is_main_sess()) set_linemode(-1);
	    } else
		tn_sendopt(DONT, *s, (char *)0);
	    tn->state = TS_DATA; break;

	case TS_WONT:
#ifdef TELOPTS
	    fprintf(stderr, "[got WONT %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_ECHO || *s == TELOPT_SGA) {
		char *topt = (*s == TELOPT_ECHO)
			? &tn->remote_echo : &tn->remote_sga;
		tn_sendopt(DONT, *s, topt);
		if(is_main_sess()) set_linemode(-1);
	    }
	    tn->state = TS_DATA; break;

	case TS_DO:
#ifdef TELOPTS
	    fprintf(stderr, "[got DO %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_NAWS || *s == TELOPT_TTYPE) {
		if(*s == TELOPT_NAWS) {
		    tn_sendopt(WILL, *s, &tn->will_naws);
		    send_term_size();
		} else
		    tn_sendopt(WILL, *s, &tn->will_ttype);
	    } else
		tn_sendopt(WONT, *s, (char *)0);
	    tn->state = TS_DATA; break;

	case TS_DONT:
#ifdef TELOPTS
	    fprintf(stderr, "[got DONT %s]\n", TELOPTSTR(*s));
#endif
	    if(*s == TELOPT_NAWS || *s == TELOPT_TTYPE) {
		char *topt = (*s == TELOPT_NAWS)
			? &tn->will_naws : &tn->will_ttype;
		tn_sendopt(WONT, *s, topt);
	    }
	    tn->state = TS_DATA; break;

	case TS_SB:
	    if(*s == IAC)
		{ tn->state = TS_SBIAC; break; }
	    lbuf_accum(tn->subopt, *s);
	    break;

	case TS_SBIAC:
	    if(*s == IAC) {
		lbuf_accum(tn->subopt, *s);
		tn->state = TS_SB; break;
	    }
	    if(*s != SE)
		to_footnote("#warning: telnet suboption not IAC SE terminated");
	    tn_suboption(tn->subopt);
	    lbuf_free(&tn->subopt);
	    tn->state = TS_DATA; break;
	}
	s++; n--;
    } while(n);
    return (char*)d - buf;
}


/*
 * set the linemode, and pick appropriate input handler
 * NOTE: should we send edline when entering charmode?
 * TODO: option linemode={auto,line,char}
 */
void set_linemode(lm)
int lm;
{
    struct telnet_data *tn = &SOCK_STATE(sockfd);
    input_proc *h = use_handler;
    if(lm < 0) lm = linemode;
    if(lm == LM_ESCAPED) {
	if(!h) h = handle_escaped;
    } else if(tn->remote_sga) {
	lm = LM_CHAR;
	h = handle_charmode;
    } else {
	lm = LM_NORMAL;
	if(tn->remote_echo) h = handle_echoless;
    }
    input_handler = h;
    if(lm != linemode) {
	linemode = lm;
	fflush(stdout);
	set_spec_keys(0);
    }
}

/*
 * send an option negotiation
 * 'what' is one of WILL, WONT, DO, DONT
 */
static void tn_sendopt(what, opt, ostate)
int what, opt;
char *ostate;
{
    static unsigned char buf[3] = { IAC, 0, 0 };
    if(ostate) {
#if (DO & 1) && (WILL & 1) && !(DONT & 1) && !(WONT & 1)
	char val = (what & 1);
#else
	char val = (what == DO || what == WILL);
#endif
	if(*ostate == val) return;	/* loop prevention rule */
	*ostate = val;
    }
    buf[1] = what; buf[2] = opt;
    send_raw(buf, 3);
#ifdef TELOPTS
    fprintf(stderr, "[sent %s %s]\n", (what == WILL) ? "WILL" :
	   (what == WONT) ? "WONT" :
	   (what == DO) ? "DO" : (what == DONT) ? "DONT" : "error",
	   TELOPTSTR(buf[2]));
#endif
}

/*
 * process suboptions.
 * right now, only terminal type is processed.
 */
static void tn_suboption(sub)
lbuf *sub;
{
    static unsigned char const sb_ttype[] =
	    { IAC, SB, TELOPT_TTYPE, TELQUAL_IS };
    static unsigned char const sb_end[] = { IAC, SE };
    if(sub->pos >= 2
	    && sub->buf[0] == (char)TELOPT_TTYPE
	    && sub->buf[1] == (char)TELQUAL_SEND) {
	/*
	 * Send our terminal type (RFC 1091)
	 * NOTE: assuming here that terminal_type is a valid name.
	 */
	char *term = terminal_type;	/* assert(terminal_type); */
	send_raw(sb_ttype, 4);
	send_binary(term, str_len(term));
	send_raw(sb_end, 2);
#ifdef TELOPTS
	fprintf(stderr, "[got SB TERMINAL TYPE SEND]\n");
	fprintf(stderr, "[sent SB TERMINAL TYPE IS %s]\n", term);
#endif
    }
}

/*
 * Send current terminal size (RFC 1073)
 */
void send_term_size()
{
    static unsigned char const sb_naws[] = { IAC, SB, TELOPT_NAWS };
    static unsigned char const sb_end[] = { IAC, SE };
    struct telnet_data *tn = &SOCK_STATE(sockfd);
    if(tn->will_naws) {
	static unsigned char buf[4];
	buf[0] = cols >> 8;
	buf[1] = cols & 0xff;
	buf[2] = lines >> 8;
	buf[3] = lines & 0xff;
	send_raw(sb_naws, 3);
	send_binary(buf, 4);
	send_raw(sb_end, 2);
#ifdef TELOPTS
	fprintf(stderr, "[sent term size: %d %d]\n", cols, lines);
#endif
    }
}

/*
 * put data in the output buffer for transmission to the remote host
 * enum { BINARY = -2, RAW = -1, DATA = 0, LINE = 1 } mode;
 */
void send_to_host(buf, len, mode)
char *buf; int len;
int mode;
{
    char *s, c;
    int i, n;
#ifndef OMIT_SESSIONS
    if(output_socket != sockfd) {	/* is there data to another socket? */
	flush_output_buffer();		/* then flush it */
	output_socket = sockfd;
    }
#endif
    for(i = output_len, s = buf, n = len; ; ) {
	if(i >= (int)sizeof(output_buffer) - 4) {
	    output_len = i;
	    flush_output_buffer();
	    i = 0;
	}
	if(!n) break;
	n--;
	c = output_buffer[i++] = *s++;
	if(c == (char)IAC && mode != -1)
	    output_buffer[i++] = c;
	else if(c == '\r' && mode >= 0 /* && !tn->will_binary */) {
	    output_buffer[i++] = 0;
	}
    }
    if(mode == 1) {
	/*
	 * TODO: allow user to choose between CR LF, CR NUL,
	 * or LF as the end of line convention.
	 */
	output_buffer[i++] = '\r';
	output_buffer[i++] = '\n';
	expect_prompt++;
    }
    output_len = i;
}

/*
 * send all buffered data to the remote host
 */
void flush_output_buffer()
{
    if(output_len) {
	if(output_socket)
	    write_socket(output_buffer, output_len);
	output_len = 0;
    }
}

/*
 * write len bytes of buffer to socket.
 * NOTE: no guarantee we don't block here indefinitely.
 */
void write_socket(buffer, len)
char *buffer; int len;
{
    char *p = buffer;
    int l = len, fd = output_socket;
    while(l) {
	int n = write(fd, p, l);
	if(n < 0) {
	    if(errno == EINTR) continue;
	    syserr("write to socket");
	}
	p += n; l -= n;
    }
}

/*
 * read a maximum of size bytes from remote host.
 * return bytes read.
 */
int read_socket(buffer, size)
char *buffer; int size;
{
    int i = read(sockfd, buffer, size);
    if(i < 0)
	syswarn("#warning: read from socket");
    return i;
}

