/*
 * term.c - terminal handling routines for cancan
 */
#include <stdio.h>
#include <string.h>
#include <signal.h>
#include <sys/ioctl.h>

#if !defined(USE_SGTTY) && !defined(USE_TERMIO)
#include <termios.h>
#endif
#ifdef USE_TERMIO
#include <termio.h>
#endif
#ifdef USE_SGTTY
#include <sgtty.h>
#endif

#ifndef VT100
int tgetent();
int tgetnum();
int tgetflag();
char *tgetstr();
char *tgoto();
#endif

#include "cancan.h"
#include "list.h"
#include "term.h"
#include "parse.h"
#include "ctypes.h"

static void gotoxy();
static void optmove();
static int regular_prompt();
static int wrap_edline();
static int edline_feed();
static void put_s();
static void put_sn();

static void set_edattr();
static void clear_attr();
static void clear_bg_attr();

static void mark_and_print();
static char *wrap_ugly();

#ifndef VT100

static int do_dch();
static int do_ich();

static int tc_extract();
static void tc_cache_gen();
#define TCS_CACHED	12	/* IC/DC precomputed with params */
#define tc_assign(tc, val) ((tc)->cost = str_copy((tc)->str, val))

/* terminal escape sequences */
tc_string tcap_cm, tcap_ce, tcap_cd, tcap_cr, tcap_le, tcap_up;
tc_string tcap_im, tcap_ei, tcap_ic, tcap_IC[TCS_CACHED];
tc_string tcap_dm, tcap_ed, tcap_dc, tcap_DC[TCS_CACHED];
tc_string tcap_me, tcap_md, tcap_mr, tcap_us, tcap_so, tcap_se;
tc_string tcap_ks, tcap_ke, tcap_kl, tcap_kr, tcap_ku, tcap_kd;
#define curgoto		tcap_cm.str
#define clreoln		tcap_ce.str
#define clreoscr	tcap_cd.str
#define begoln		tcap_cr.str
#define leftcur		tcap_le.str
#define upcur		tcap_up.str
#define modenorm	tcap_me.str
#define modebold	tcap_md.str
#define modeinv		tcap_mr.str
#define modeuline	tcap_us.str
#define modestandon	tcap_so.str
#define modestandoff	tcap_se.str
#define kpadstart	tcap_ks.str
#define kpadend		tcap_ke.str
#define gotocost	tcap_cm.cost
#define len_begoln	tcap_cr.cost
#define len_leftcur	tcap_le.cost
#define len_upcur	tcap_up.cost
#define len_clreoln	tcap_ce.cost

static int insertmode_cost, insertfinish;
static int deletemode_cost;

#ifndef metakey
int metakey;		/* terminal has meta key setting high bit */
#endif
int wrapglitch;		/* strange wrap behaviour (a la vt100) */

char *terminal_type;

#else	/* hard-coded vt100 features if no termcap: */

tc_string curgoto = "\033[%d;%dH", clreoln = "\033[K", clreoscr = "\033[J";
tc_string begoln = "\r", leftcur = "\b", upcur = "\033[A";

tc_string modenorm = "\033[m";
tc_string modebold = "\033[1m", modeuline = "\033[4m", modeinv = "\033[7m";
tc_string modestandon = "\033[7m", modestandoff = "\033[m";

tc_string kpadstart = "\033[?1h\033=", kpadend = "\033[?1l\033>";
tc_string key_left = "\033OD", key_right = "\033OC";
tc_string key_up = "\033OA", key_down = "\033OB";

#define gotocost	8
#define len_clreoln	3
#define len_begoln	1
#define len_leftcur	1
#define len_upcur	3

#define wrapglitch	1
#define insertfinish	0

char *terminal_type = "vt100";

#endif /* VT100 */

#ifndef LINES
#define LINES 24
#endif
#ifndef COLUMNS
#define COLUMNS 80
#endif

int lines = LINES;	/* screen size */
int cols = COLUMNS;

int lines_1 = LINES - 1;/* last screen line */
int cols_1 = COLUMNS;	/* either cols, or cols-1 if no wrapglitch */

int line0 = 0;		/* screen line where the input line starts */
int col0 = 0;		/* input line offset (= length of prompt) */

/* input line highlighting */
int edattr = NOATTRCODE;/* input line attributes */
int edattr_active;
char edattrbeg[CAPLEN];	/* start input line attributes */
char edattrend[CAPLEN]; /* end input line attributes */

#ifdef WITH_STDCOLOUR
char mode_std[CAPLEN];
#endif

#ifdef XTERM_JS_BUG
static int leave_bg_off;
#endif


#ifndef VT100
/*
 * read termcap definitions
 * NOTE: We should check if term has "in" insert_null_glitch (rare).
 */
void init_termcap()
{
    struct tc_init_node {
	char id[3], critical;
	tc_string *tcap;
    };
    static struct tc_init_node const tc_init[] = {
	{ "cm", 1, &tcap_cm },
	{ "ce", 1, &tcap_ce }, { "cd", 1, &tcap_cd },
	{ "cr", 0, &tcap_cr },
	{ "le", 0, &tcap_le }, { "up", 0, &tcap_up },
	{ "im", 0, &tcap_im }, { "ei", 0, &tcap_ei },
	{ "ic", 0, &tcap_ic }, { "IC", 0,  tcap_IC },
	{ "dm", 0, &tcap_dm }, { "ed", 0, &tcap_ed },
	{ "dc", 0, &tcap_dc }, { "DC", 0,  tcap_DC },
	{ "me", 0, &tcap_me }, { "md", 0, &tcap_md },
	{ "mr", 0, &tcap_mr }, { "us", 0, &tcap_us },
	{ "so", 0, &tcap_so }, { "se", 0, &tcap_se },
	{ "ks", 0, &tcap_ks }, { "ke", 0, &tcap_ke },
	{ "kl", 0, &tcap_kl }, { "kr", 0, &tcap_kr },
	{ "ku", 0, &tcap_ku }, { "kd", 0, &tcap_kd },
	{ "",   0, 0 }
    };
    char tcbuf[2048];			/* by convention, this is enough */
    const struct tc_init_node *np;
    char *term = get_env("TERM");
    int i;

    if(!term || !(i = tgetent(tcbuf, term))) {
	if(term) fprintf(stderr,
		    "There is no entry for '%s' in the terminal data base.\n",
		    term);
	fprintf(stderr, "Please set your TERM environment variable.\n");
	exit(1);
    }
    if(i < 0) {
	syswarn("Couldn't read termcap database");
	exit(1);
    }
    terminal_type = str_dup(term);	/* record it for later use */

    for(np = tc_init; np->id[0]; np++) {
	tc_string *tc = np->tcap;
	tc->cost = tc_extract(np->id, tc->str);
	if(!tc->cost && np->critical) {
	    fprintf(stderr,
		    "Your terminal is not powerful enough, missing '%s'.\n",
		    np->id);
	    exit(1);
	}
    }
    if(!tcap_cr.cost) tc_assign(&tcap_cr, "\r");
    if(!tcap_le.cost) tc_assign(&tcap_le, "\b");
    /* check_winsize() will override these later: */
    if((i = tgetnum("li")) > 0) lines_1 = (lines = i) - 1;
    if((i = tgetnum("co")) > 1) cols_1 = cols = i;
    gotocost = str_len(tgoto(curgoto, cols_1, lines_1));
#ifndef metakey
    metakey = tgetflag("km");
#endif
    wrapglitch = tgetflag("xn");
    if(!wrapglitch)
	{ cols_1--; insertfinish = gotocost + len_clreoln; }
    tc_cache_gen(tcap_IC);
    tc_cache_gen(tcap_DC);
    insertmode_cost = tcap_im.cost + tcap_ei.cost;
#ifndef TERMCAP_ORTHODOX
    if(tcap_ic.cost) {
	tc_assign(tcap_IC + 1, tcap_ic.str);
	if(insertmode_cost) { tcap_ic.cost = 0; tcap_ic.str[0] = 0; }
    }
#endif
    if(tcap_dc.cost) {
	deletemode_cost = tcap_dm.cost + tcap_ed.cost;
	if(!deletemode_cost) tc_assign(tcap_DC + 1, tcap_dc.str);
    }
}

/*
 * extract termcap 'cap' and copy it to buf.
 * return the lenght of the extracted string.
 */
static int tc_extract(cap, buf)
char *cap, *buf;
{
    char *d = buf, *s = tgetstr(cap, &buf);
    int len = 0;
    /*
     * NOTE: We assume terminals require no padding nowadays, so we
     * remove padding information. Perhaps we could use it in cost
     * calculation, though.
     * NOTE: We don't check for buf overflow.
     */
    if(s) {
	s += digit_span(s);
	if(*s == '.' && *++s) ++s;
	if(*s == '*') s++;
	/* if(*s == '/') ; else */
	for( ; *s; s++)
	    if(s[0] == '$' && s[1] == '<') {
		if(!(s = strchr(s, '>'))) break;
	    } else {
		len++; *d++ = *s;
	    }
    }
    *d = 0;
    return len;
}

static void tc_cache_gen(tcs)
tc_string *tcs;
{
    if(tcs->cost) {
	int i;
	for(i = TCS_CACHED; --i; )
	    tc_assign(tcs + i, tgoto(tcs->str, i, i));
	tcs->cost = (tcs + TCS_CACHED - 1)->cost;	/* estimate */
    }
}
#endif /* !defined(VT100) */

/*
 * create the default keypad bindings
 */
void keypad_bindings()
{
    /*
     * NOTE: termcap doesn't have sequences for the numeric keypad,
     * so we just assume they are the same as for a vt100.
     */
    struct keypad_init_node {
	char label[4], seq[4], cmd[6];
    };
    static struct keypad_init_node const keypad_init[] = {
	{ "kp2", "\033Or", "s" },
	{ "kp3", "\033Os", "d" },
	{ "kp4", "\033Ot", "w" },
	{ "kp5", "\033Ou", "exits" },
	{ "kp6", "\033Ov", "e" },
	{ "kp8", "\033Ox", "n" },
	{ "kp9", "\033Oy", "u" },
	{ "", "", "" }
    };
    const struct keypad_init_node *p = keypad_init;
    do {
	keynode *k = *add_keynode(p->label, p->seq, p->cmd);
	k->keydef.func = run_command;
    } while((++p)->label[0]);
}

/*
 * initialize the key binding list
 */
void default_bindings()
{
    /*
     * NOTE: in the unlikely event that two keys (down, up, etc.) have
     * colliding sequences, only the first one is added to keymap tree.
     */
    const struct defbind_node *p = defbinds;
    do if(p->seq[0]) {
	keynode *k = *add_keynode(p->label, p->seq, p->name);
	k->keydef.func = p->func;
	k->type = p->type;
    } while((++p)->label[0]);
}


/*
 * position the cursor using absolute coordinates
 */
static void gotoxy(col, line)
int col, line;
{
#ifndef VT100
    put_s(tgoto(curgoto, col, line));
#else
    printf(curgoto, line + 1, col + 1);
#endif
}

/*
 * optimized cursor movement
 * from (fromline, fromcol) to (toline, tocol)
 * if tocol > 0, (toline, tocol) must reside on editline.
 */
static void optmove(fromline, fromcol, toline, tocol)
int fromline, fromcol, toline, tocol;
{
    char buf[CAPLEN], *cp = buf;
    int i, credit = gotocost;
    /*
     * First, move vertically to the correct line, then horizontally
     * to the right column. If this turns out to be fewer characters
     * than a direct cursor positioning (gotoxy), use it.
     */
    for(;;) {
	if((i = toline - fromline) < 0) {
	    if(!len_upcur || (credit -= -i * len_upcur) < 0) break;
	    do { cp += str_copy(cp, upcur); } while(++i);
	} else if(i) {
	    /* lf's are mapped into crlf on output */
	    if((credit -= 2 * i) < 0) break;
	    do { *cp++ = '\n'; } while(--i);
	    fromcol = 0;
	}
	if(fromcol >= cols) break;	/* we are in a glitch state */
	if((i = tocol - fromcol) < 0) {
	    int dist = -i * len_leftcur;
	    if(dist <= len_begoln + tocol) {
		if((credit -= dist) < 0) break;
		do { cp += str_copy(cp, leftcur); } while(++i);
	    } else {
		/* assume we can redraw prompt if we need to */
		if((credit -= len_begoln) < 0) break;
		cp += str_copy(cp, begoln);
		fromcol = 0;  i = tocol;
	    }
	}
	if(i) {
	    if((credit -= i) < 0) break;
	    if(fromcol < col0 && toline == line0) {
		if(edattrbeg[0]) break; /* gotoxy if hiliting used */
		if(!regular_prompt()) break;
		/* TODO: use gotoxy if prompt was marked. */
		do { *cp++ = promptstr[fromcol]; } while(++fromcol < col0);
	    }
	    if(fromcol < tocol) {
		set_edattr();
		i = (toline - line0) * cols_1 - col0;
		fromcol += i; tocol += i;
		do { *cp++ = edbuf[fromcol]; } while(++fromcol < tocol);
	    }
	}
	*cp = 0;
	put_s(buf);
	return;
    }
    gotoxy(tocol, toline);
}

/*
 * change the position on input line (optmove there, and set pos)
 */
void moveto(new_pos)
int new_pos;
{
    /*
     * FEATURE: the line we are moving to might be less than 0, or greater
     * than lines_1, if the display is too small to hold the whole editline.
     * In that case, the input line should be (partially) redrawn.
     */
#if 0	/* optimize for the most likely case */
    int cc = CURCOL(pos), cl = CURLINE(pos);
    if(cc && pos == new_pos + 1)
	put_s(leftcur);
    else
	optmove(cl, cc, CURLINE(new_pos), CURCOL(new_pos));
#else
    optmove(CURLINE(pos), CURCOL(pos), CURLINE(new_pos), CURCOL(new_pos));
#endif
    pos = new_pos;
}

/*
 * move to the beginning of last line
 */
void lastline()
{
    /* NOTE: could use tcap "ll" here */
    gotoxy(0, line0 = lines_1);
}

/*
 * after parse(), make sure we know on which line we are
 */
void check_line0()
{
    if(line0 != lines_1) lastline();
}

/*
 * print a newline and advance line0.
 */
void newline()
{
    putc('\n', stdout);
    if(line0 < lines_1) line0++;
#ifdef XTERM_JS_BUG
    else leave_bg_off = 1;
#endif
}

/*
 * display or clear a one-line footnote at the bottom of the screen
 * TODO: check msg length, limit it by cols_1.  Use terminal status
 * line if there is one (tcaps "hs", "ts", "fs", "ws", "ds").
 */
void display_footnote()
{
    char *msg = footnote;
    int to_c = CURCOL(pos);
    int to_l = CURLINE(pos);
    optmove(to_l, to_c, lines_1, 0);
    if(msg) {
	clear_attr();
	if(CURLINE(edlen) >= lines_1)
	    { --line0; --to_l; put_s("\n"); }
	put_s(msg);
	/* fflush(stdout); sleep(1); */
    } else
	put_s(clreoln);
    gotoxy(to_c, to_l);
}

/*
 * execute string as if typed
 */
void run_command(cmd)
char *cmd;
{
    int len = str_len(cmd);
    to_capfile(cmd, len, 1);
    clear_input();
    line0 = wrap_edline(cmd, len);
    clear_attr();
    newline();
    parse(cmd, 1, (struct param *)0);
    check_line0();
    redraw_all();
}

/*
 * skip over the edit line, set new line0 and turn edattribs off
 */
void skip_edline()
{
    int p = pos;
    int cl = CURLINE(p);
    int nl = p - CURCOL(p);
    clear_attr();
    line0 = cl;
    do {
	newline();
	nl += cols_1;
    } while(nl < edlen);
}

/*
 * wrap out an [edit] line, leave cursor at the end of it
 */
static int wrap_edline(line, len)
char *line; int len;
{
    int p = 0;
    int cl = line0;
    int nl = -col0;
    for( ; ; ) {
	set_edattr();
	nl += cols_1;
	if(nl > len) break;
	put_sn(line + p, nl - p); p = nl;
	cl = edline_feed(cl);
    }
    put_sn(line + p, len - p);
    return cl;
}

/*
 * like newline(), but for use with edline operations
 */
static int edline_feed(ln)
int ln;
{
    if(++ln > lines_1) {
	ln = lines_1;
	--line0;
	clear_bg_attr();
#ifdef XTERM_JS_BUG
	leave_bg_off = 1;
#endif
    }
    put_s("\n");
    return ln;
}

/*
 * clear current input line (but not the prompt)
 * cursor is left right after the prompt
 */
void clear_input()
{
    if(input_handler && (*input_handler)(DO_CLEAR_INPUT))
	return;
    clear_attr();
    if(edlen) {
	int from_c = CURCOL(pos);
	int from_l = CURLINE(pos);
	if(line0 < lines_1) {
	    /* NOTE: plain clreoscr would suffice when col0 == 0 */
	    optmove(from_l, from_c, line0 + 1, 0);
	    put_s(clreoscr);
	    from_c = 0;  from_l = line0 + 1;
	}
	optmove(from_l, from_c, line0, col0);
	put_s(clreoln);
    }
}

/*
 * Redraw the input line and put the cursor at the current position.
 * The cursor is assumed to be directly after the prompt.
 */
void redraw_input()
{
    if(input_handler && (*input_handler)(DO_REDRAW_INPUT))
	return;
    if(edlen) {
	int len = edlen;
	wrap_edline(edbuf, len);
	if(len != pos) {
	    int t = pos;
	    pos = len;
	    moveto(t);
	}
#ifdef XTERM_JS_BUG
	if(leave_bg_off) clear_bg_attr();
#endif
    }
}

/*
 * clear the prompt if there is one
 */
void clear_prompt()
{
    if(input_handler && (*input_handler)(DO_CLEAR_PROMPT))
	return;
    if(col0) {
	put_s(begoln);
	put_s(clreoln);
    }
}

/*
 * redisplay the prompt if needed
 */
void redraw_prompt()
{
    if(input_handler && (*input_handler)(DO_REDRAW_PROMPT))
	return;
    if(!expect_prompt && col0 && regular_prompt()) {
	mark_and_print(promptstr, col0);
    } else
	col0 = 0;
}

/*
 * clear both prompt and input line
 */
void clear_all()
{
    gotoxy(0, line0);
    clear_attr();
    put_s(clreoscr);
}

/*
 * redraw both prompt and input line
 */
void redraw_all()
{
    redraw_prompt();
    redraw_input();
}

/*
 * return false if no prompt or it contains control chars
 */
static int regular_prompt()
{
    int c, n = -1;
    do {
	c = (unsigned char)promptstr[++n];
    } while(isprint(c));
    return c ? 0 : n;
}


#ifndef VT100
/*
 * Use parm_dch or dch when available and cheaper than redraw.
 */
static int do_dch(maxcost, n, clr_bg)
int maxcost, n, clr_bg;
{
    tc_string *tc;
    int cost;
    if(n <= 0)
	return 0;
#ifdef XTERM_ICH_GLITCH
    if(BACKGROUND(edattr) != ATTR_NONE)
	return 0;
#endif
    tc = tcap_DC + (n < TCS_CACHED ? n : 0);
    if((cost = tc->cost)) {
	if(cost < maxcost) {
	    if(clr_bg) clear_bg_attr();
	    put_s(n < TCS_CACHED ? tc->str : tgoto(tc->str, n, n));
	    return 1;
	}
    } else if((cost = deletemode_cost + n * tcap_dc.cost)) {
	if(cost < maxcost) {
	    if(clr_bg) clear_bg_attr();
	    put_s(tcap_dm.str);
	    do {
		put_s(tcap_dc.str);
	    } while(--n);
	    put_s(tcap_ed.str);
	    return 1;
	}
    }
    return 0;
}

/*
 * Use parm_ich, ich or smir/rmir when available and cheaper than redraw.
 */
static int do_ich(maxcost, n, str)
int maxcost, n; char *str;
{
    tc_string *tc;
    int cost;
    if(n <= 0)
	return 0;
#ifdef XTERM_ICH_GLITCH
    if(BACKGROUND(edattr) != ATTR_NONE)
	return 0;
#endif
    tc = tcap_IC + (n < TCS_CACHED ? n : 0);
    if((cost = tc->cost)) {
	if(cost < maxcost) {
	    put_s(n < TCS_CACHED ? tc->str : tgoto(tc->str, n, n));
	    put_sn(str, n);
	    return 1;
	}
    } else if((cost = insertmode_cost + n * tcap_ic.cost)) {
	if(cost < maxcost) {
	    put_s(tcap_im.str);
	    do {
		put_s(tcap_ic.str);
		put_sn(str++, 1);
	    } while(--n);
	    put_s(tcap_ei.str);
	    return 1;
	}
    }
    return 0;
}
#endif

#define BUFDEL(buf, blen, n)				\
    do {						\
	int i, k = (n), l = (blen) - k;			\
	char *b = (buf);				\
	for(i = 0; i < l; i++)				\
	    b[i] = b[i + k];				\
    } while(0)

#define BUFINS(buf, blen, str, slen)			\
    do {						\
	int i, bl = (blen), sl = (slen);		\
	char *b = (buf), *s = (str);			\
	for(i = bl; --i >= 0; b[i + sl] = b[i]);	\
	for(i = sl; --i >= 0; b[i] = s[i]);		\
    } while(0)

/*
 * delete n characters at current position (the position is unchanged)
 * assert(n <= edlen - pos)
 */
void delete_chars(n)
int n;
{
    int p = pos;
    int nl, cl;		/* next line's starting pos, current line */
    int r_cost, d_cost;
    int oldlen;
    BUFDEL(edbuf + p, edlen - p + 1, n);
    oldlen = edlen; edlen -= n;
    nl = p - CURCOL(p);
    cl = CURLINE(p);
    /* -gotocost: no goto after delete in one-line operation */
    for(d_cost = -gotocost; ; d_cost = 0) {
	d_cost -= len_clreoln;	/* delete is advantageous on last line */
	r_cost = edlen;
	nl += cols_1;
	if(oldlen > nl) {	/* need process next line too */
	    d_cost = edlen - nl;
	    if(d_cost > 0)	/* new edline continues to next line */
		d_cost = 0, r_cost = nl;
	    d_cost += gotocost + n;
	}
	r_cost -= p;
	if(r_cost) {
#ifndef VT100
	    if(do_dch(r_cost - d_cost, n, (oldlen <= nl))) {
		if(oldlen <= nl) break;
		gotoxy(cols_1 - n, cl);
		r_cost += p;  r_cost -= (p = nl - n);
	    }
#endif
	    set_edattr();
	    put_sn(edbuf + p, r_cost); p += r_cost;
	}
	if(p < nl) {		/* clear to eol */
	    clear_bg_attr();
	    if(n == 1 && p < nl - 1) {
		p++; put_s(" ");
	    } else
		put_s(clreoln);
	    if(oldlen <= nl) break;
	}
	put_s("\n"); cl++; p = nl;
	if(edlen <= nl) {	/* clear to eos */
	    clear_bg_attr();
	    put_s(clreoscr);
	    break;
	}
    }
    if(p != pos) {
	int t = pos;
	pos = p;
	moveto(t);
    }
}

/*
 * insert n characters at current position. The position is set to after
 * the inserted characters.
 */
void insert_chars(str, n)
char *str; int n;
{
    int p = pos;
    int nl, cl;		/* next line's starting pos, current line */
    BUFINS(edbuf + p, edlen - p + 1, str, n);
    edlen += n; pos += n;
    nl = p - CURCOL(p);
    cl = CURLINE(p);
    do {
	int i_cost = n;
	int r_cost = edlen;
	nl += cols_1;
	if(r_cost > nl)
	    r_cost = nl, i_cost += insertfinish;
	r_cost -= p;
	set_edattr();
#ifndef VT100
	if(do_ich(r_cost - i_cost, n, edbuf + p)) {
	    p += n;
	    if(insertfinish && edlen > nl) {
		gotoxy(cols_1, cl);
		put_s(clreoln);
	    }
	} else
#endif
	{
	     put_sn(edbuf + p, r_cost); p += r_cost;
	}
	if(edlen < nl) break;
	cl = edline_feed(cl); p = nl;
    } while(edlen > nl);
    if(p != pos) {
	int t = pos;
	pos = p;
	moveto(t);
    }
#ifdef XTERM_JS_BUG
    if(leave_bg_off) clear_bg_attr();
#endif
}

/*
 * insert one character. This is optimized for the most common
 * case of (pos == edlen).
 */
void insert_one_char(s)
char *s;
{
    if(pos >= edlen) {
	int p = replace_char(*s);
	edbuf[edlen = p] = 0;
    } else
	insert_chars(s, 1);
}

/*
 * print a char on current position (overwrite).
 * return new pos (advanced by one).
 */
int replace_char(c)
char c;
{
    int p = pos;
    set_edattr();
    putc(c, stdout);
    edbuf[p] = c;
    pos = ++p;
    if(!CURCOL(p)) {
	edline_feed(CURLINE(p) - 1);
#ifdef XTERM_JS_BUG
	if(leave_bg_off) clear_bg_attr();
#endif
    }
    return p;
}


/*
 * print a (short) null terminated string
 */
static void put_s(str)
char *str;
{
#if !defined(putc) || defined(putc_unlocked)
    fputs(str, stdout);
#else
    char c, *s = str;
    FILE *f = stdout;
    while((c = *s++)) putc(c, f);
#endif
}

/*
 * print a string with length n
 */
static void put_sn(str, len)
char *str; int len;
{
    fwrite(str, 1, len, stdout);
}

/*
 * NOTE: coloured background handling can be a pain.
 * We need to turn bg. off whenever new lines are created with scrolling, or
 * blanks shifted in when delchar is used. (New space is usually created with
 * current bg.)  See README.xterm for description of some bg. colour related
 * bugs in xterm.
 * NOTE: we should do attr ==> attr switching, not simply attr on/off.
 */

static void set_edattr()
{
    if(!edattr_active) {
	edattr_active = 1;
	put_s(edattrbeg);
    }
}

static void clear_attr()
{
    if(edattr_active) {
	edattr_active = 0;
	put_s(edattrend);
    }
}

static void clear_bg_attr()
{
#ifdef XTERM_JS_BUG
    leave_bg_off = 0;
#endif
    if(BACKGROUND(edattr) != ATTR_NONE)
	clear_attr();
}

/*
 * put escape sequences to turn on/off an attribute into given buffers
 */
void attr_string(attrcode, begin, end)
int attrcode; char *begin, *end;
{
    const char *startseq, *endseq;
    attr_token fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);
    if(fg < ATTR_COLOURS || bg < ATTR_COLOURS) {
	char *s = begin;
	*s = '\033'; *++s = '[';
	if(fg < ATTR_COLOURS)
	    { *++s = '3'; *++s = i_to_digit(fg); *++s = ';'; }
	if(bg < ATTR_COLOURS)
	    { *++s = '4'; *++s = i_to_digit(bg); ++s; }
	*s = 'm'; *++s = 0;
	endseq = modenorm;
#ifdef WITH_STDCOLOUR
	if(mode_std[0])
	    endseq = mode_std;
#endif
    } else {
	*begin = *end = '\0';
	if(fg == ATTR_UNDERLINE)
	    startseq = modeuline;
	else if(fg == ATTR_BOLD)
	    startseq = modebold;
	else if(fg == ATTR_INVERSE || fg == ATTR_REVERSE)
	    startseq = modeinv;
	else return;
	endseq = modenorm;
	if(!*startseq) {
	    startseq = modestandon;
	    endseq = modestandoff;
	}
	strcpy(begin, startseq);
    }
    strcpy(end, endseq);
}

#ifdef WITH_STDCOLOUR
int set_stdcolour(attrcode)
int attrcode;
{
    if(attrcode != NOATTRCODE) {
	attr_token fg = FOREGROUND(attrcode), bg = BACKGROUND(attrcode);
	if(!(fg < ATTR_COLOURS && bg < ATTR_COLOURS))
	    return 0;
	sprintf(mode_std, "\033[3%d;4%dm", fg, bg);
	put_s(mode_std);
    } else {
	mode_std[0] = 0;
	put_s(modenorm);
    }
    attr_string(edattr, edattrbeg, edattrend);
    return 1;
}
#endif


/*
 * NOTE: An altogether different pattern matching scheme should be devised--
 * probably one with marker info (and actions?) embedded in the pattern.
 *
 * Marking is performed regardless of possible attribute changes in the text.
 * Cancan uses attributes, but doesn't really handle them. For instance, try
 * "#mark \ =bold", and see what output is produced for long space sequences.
 */

static void pm_preprocess();
static int pm_search();
static int pm_match();

static unsigned char pm_alpha[256], pm_delta[LINELEN];

/*
 * NOTE: in anticipation of trying several patterns against the text,
 * some preprocessing is done on the latter. (hashing, sort of).
 * alpha[c] is an index (mod 256) to first char c in text, or anything
 * at all if no occurrences are found. (256 - delta[pos]) is the number
 * of positions to skip forward when match at pos fails.
 * assert(MAX_UCHAR + 1 == 256);
 */
static void pm_preprocess(str, len)
char *str; int len;
{
    register char *s = str;
    register int i = len;
    while(--i >= 0) {
	register int c = (unsigned char)s[i];
	pm_delta[i] = i - pm_alpha[c];
	pm_alpha[c] = i;
    }
}

/*
 * search for marks in a string str of length len, starting from pos i.
 * return the position of match (or -1 for none), and fill in *match.
 *
 * NOTE: perhaps we ought to match a marker starting/ending with
 * a !isword() char even if it's not preceded/followed by one.
 * NOTE: we may test isword(str[len]), so this char counts too.
 */
static int pm_search(str, i, len, match)
char *str; int i; int len;
marknode **match;
{
    marknode *np;
    int mpos = -1;

    for(np = markers; np; np = np->next) {
	register int spos, scan = len - np->plen;
	register char *s, *p;

	if(scan < 0) continue;
	if((unsigned)scan >= (unsigned)mpos) {
	    /* don't scan past previous match position */
	    scan = mpos;
	    /* only longer pattern is tried at prev. mpos */
	    if(np->plen <= (*match)->plen) scan--;
	}
	/* if(scan < i) continue; */

	p = np->pattern;
	spos = i + (unsigned char)(pm_alpha[(unsigned char)*p] - i);
	for( ; spos <= scan; spos += (unsigned char)-1 + 1) {
	    s = str + spos;
	    if(*s != *p) continue;
	    if((!spos || !isword(s[-1])) && !isword(s[np->plen])) {
		do p++, s++; while(*p && *s == *p);
		if(!*p) {			/* a match! */
		    mpos = spos; *match = np;
		    break;
		}
		p = np->pattern;
	    }
	    spos -= pm_delta[spos];
	}
    }
    return mpos;
}

/*
 * while scanning text from pos. i, we have found a match at mpos.
 * now, print out the processed portion of the text, and modify the
 * alpha[] table. return new pos. to start scanning from.
 */
static int pm_match(str, i, mpos, m)
char *str; int i, mpos; marknode *m;
{
    static char begin[CAPLEN], end[CAPLEN];

    attr_string(m->attrcode, begin, end);
    put_sn(str + i, mpos - i);
    put_s(begin);
    put_sn(m->pattern, m->plen);
    put_s(end);

    for(mpos += m->plen; i < mpos; i++)
	pm_alpha[(unsigned char)str[i]] -= pm_delta[i];
    return i;
}

/*
 * print line, scanning for marked text
 */
static void mark_and_print(str, len)
char *str; int len;
{
    if(markers) {
	marknode *match;
	int mpos, i;
	pm_preprocess(str, len);
	for(i = 0; ; ) {
	    mpos = pm_search(str, i, len, &match);
	    if(mpos < 0) break;
	    i = pm_match(str, i, mpos, match);
	    if(i >= len) return;
	}
	len -= i; str += i;
    }
    put_sn(str, len);
}


/*
 * NOTE: We do not perform any complex terminal emulation. But we want to
 * keep track of the last wrapped line (the prompt). We are able to do so
 * if the line is made up of printable chars. Otherwise, we assume that:
 *   a) the input only contains harmless control sequences, if any,
 *   b) there is no incomplete sequence at the end of the line, and
 *   c) that we don't mark text within control sequence by accident.
 *
 * TODO: wrap_out() needs to remember its state, since we may receive text
 * in multiple fragments. enum { BOL, WORD, DELIM } might be sufficient.
 * Then we can have mark_and_print call wrap_out, and allow substitutions
 * like "#mark foo=>foobar".  A scan for word boundaries before the search
 * could improve efficiency and prevent accidents with esc sequences (which
 * are considered words in this context).
 */

char *wrap_out(str, len)
char *str; int len;
{
    /* assert(str[len] == '\0'); */
    if(print_span(str) == len)			/* all printable chars ... */
	if(col0 + len <= cols_1) {		/* of suitable length ... */
	    col0 = col0 + len;
	    mark_and_print(str, len);		/* ... ain't life easy! */
	    return 0;
	}
    return wrap_ugly(str, len);
}

/*
 * write string to stdout, wrapping words to the next line.
 * return pointer to the last wrapped line, or NULL.
 *
 * FEATURE: a feeble attempt is made to handle the ansi esc sequences,
 * so that the attribute changes will probably look ok [in prompt].
 * FEATURE: right now, a marker is not recognized if it gets wrapped.
 * NOTE: perhaps we should (optionally) expand tabs.
 * NOTE: this proc needs to be rewritten (as an fsm perhaps).
 */
static char *wrap_ugly(str, len)
char *str; int len;
{
    char *tail = 0, *end = str + len;
    int cc = col0, width = cols_1;
    unsigned char *wp, *s = (unsigned char *)str;
    for(wp = 0; ; s++) {
	if(isgraph(*s)) {
	    if(++cc <= width) continue;
	} else switch(*s) {
	    default:
		continue;
	    case '\033':
		if(*++s == '[')
		    while(s[1] >= ' ' && *++s < '@');
		if(!*s) s--;		/* hmmm */
		continue;
	    case '\b':
		if(--cc > 0) continue;
	    case '\r':
	    case '\n':	/* this should also increment line0 */
		cc = 0;
		tail = (char *)s + 1;
	    case '\013':	/* these should also increment line0 */
	    case '\014':
		wp = 0; continue;
	    case '\t':
		/* tab to screen end might be glitchy, force wrap */
		if((cc | 7) + 1 < width) {
		    cc = (cc | 7) + 1;
		    wp = s; continue;
		}
		wp = 0; break;
	    case ' ':
		if(cc < width) {
		    cc++;
		    wp = s; continue;
		}
		wp = 0; break;
	    case '\0':
		if((char *)s != end)
		    continue;
		wp = 0; break;
	}
	if(wp && OPTION(OPT_WORDWRAP)) s = wp + 1;
	mark_and_print(str, (char *)s - str);
	if(*s == '\t' && cc < width) {
	    /* skip if tab crosses end of the line */
	    cc = width;
	    s++;
	}
	if((char *)s == end) break;
	cc = 0;  newline();
	tail = str = (char *)s;
	wp = 0; --s;
    }
    col0 = cc;
    return tail;
}


/*
 * our trivial builtin pager stuff
 */

static char *wrap_pager_line();
static int handle_pager();
static void delete_pagernode();

static defnode *pager_head, *pager_tail;	/* list of pager lines */

static char *wrap_pager_line(str)
char *str;
{
    char *s = str;
    int i = cols_1;
    while(*s && *s != '\n' && (++s, --i));
    put_sn(str, s - str);
    newline();
    if(OPTION(OPT_PAGERWRAP)) {
	if(*s == '\n') s++;
	if(!*s) s = 0;
    } else s = 0;
    return s;
}

/*
 * TODO: should not discard lines when redrawing the page.
 * Add prompt (bottom/total) line counter. Handle 'b' back,
 * 'r' redraw, maybe even arrow keys. Use alternate screen
 * if that's available and desired. And keep it simple! :^)
 */
static int handle_pager(c)
int c;
{
    if(c == DO_CLEAR_PROMPT || c == DO_REDRAW_PROMPT)
	return 1;	/* override normal prompt */
    do {
	int n_lines = lines_1;
	defnode *p;
	if(c != DO_REDRAW_INPUT) {
	    put_s(begoln);
	    put_s(clreoln);
	    if(c == '\r' || c == '\n')
		n_lines = 1;
	    else if(c != ' ')
		break;
	    c = ' ';	/* so that we return 1 */
	}
	while((p = pager_head)) {
	    if(n_lines) {
		char *s = p->key;
		do {
		    s = wrap_pager_line(s);
		    n_lines--;
		} while(s && n_lines);
		if(!s) {
		    delete_pagernode();
		    continue;
		} else {
		    char *d = p->key;
		    while((*d++ = *s++));
		}
	    }
	    printf("%s--More--%s", modestandon, modestandoff);
	    return 1;
	}
    } while(0);
    while(pager_head)
	delete_pagernode();
    set_input_handler((input_proc *)0);
    redraw_prompt();	/* restore normal prompt */
    if(c != DO_CLEAR_INPUT)
	redraw_input();	/* ... and input line */
    return c < 0 || isprint(c);
}

/*
 * add a line to the end of pager text
 * to_pager() should be called only when
 * both prompt and input are cleared.
 */
void to_pager(str)
char *str;
{
    defnode *nd = new_node((int)sizeof(defnode), str);
    nd->next = 0;
    if(!pager_head)
	pager_head = nd;
    else
	pager_tail->next = nd;
    pager_tail = nd;
    set_input_handler(handle_pager);
    insertbuf[0] = 0;		/* XXX */
}

static void delete_pagernode()
{
    defnode *p = pager_head;
    pager_head = p->next;
    free_node(p);
}


/*
 * low level tty handling routines
 * use termios unless requested otherwise
 */

#if !defined(USE_SGTTY) && !defined(USE_TERMIO)
typedef struct termios termiostruct;

#ifdef TCSETS	/* direct ioctl */
/* NOTE: ought to check for EINTR. cast b to (char*)? */
#define term_get_attr(fd, b) (ioctl(fd, TCGETS, (b)) == 0)
#define term_set_attr(fd, b) (ioctl(fd, TCSETS, (b)) == 0)

#else		/* POSIX functions */
#define term_get_attr(fd, b) (tcgetattr(fd, (b)) == 0)
#define term_set_attr(fd, b) (tcsetattr(fd, TCSANOW, (b)) == 0)
#endif
#endif

#ifdef USE_TERMIO
typedef struct termio termiostruct;
/* NOTE: ought to check for EINTR. cast b to (char*)? */
#define term_get_attr(fd, b) (ioctl(fd, TCGETA, (b)) == 0)
#define term_set_attr(fd, b) (ioctl(fd, TCSETA, (b)) == 0)
#endif

#ifdef USE_SGTTY
typedef struct {
    struct sgttyb sg;
    struct tchars tc;
    struct ltchars ltc;
} termiostruct;
#define term_get_attr(fd, b)				\
	(  ioctl(fd, TIOCGETP, &(b)->sg) == 0		\
	&& ioctl(fd, TIOCGETC, &(b)->tc) == 0		\
	&& ioctl(fd, TIOCGLTC, &(b)->ltc) == 0)
#define term_set_attr(fd, b)				\
	(  ioctl(fd, TIOCSETP, &(b)->sg) == 0		\
	&& ioctl(fd, TIOCSETC, &(b)->tc) == 0		\
	&& ioctl(fd, TIOCSLTC, &(b)->ltc) == 0)
#endif

static int tty_is_set;
static termiostruct ttybsave;

/*
 * Save the original state, disable echo and special keys.
 */
void set_terminal()
{
    if(term_get_attr(0, &ttybsave))
	tty_is_set = 1;
    set_spec_keys(0);
    signal(SIGINT, intrsig);
    put_s(kpadstart);
}

/*
 * Reset the terminal to its original state
 */
void reset_terminal()
{
    signal(SIGINT, SIG_IGN);
    if(tty_is_set)
	if(!term_set_attr(0, &ttybsave)) /* hm */;
    clear_attr();
    put_s(kpadend);
}

/*
 * set raw/cooked mode depending on the current linemode
 * enable/disable special keys
 */
void set_spec_keys(k)
int k;
{
    if(tty_is_set) {
	termiostruct ttyb;
	ttyb = ttybsave;		/* struct assign */
#ifndef USE_SGTTY
	ttyb.c_cc[VTIME] = 0;
	ttyb.c_cc[VMIN] = 1;
	ttyb.c_iflag &= ~(INLCR | ICRNL);
#ifdef DO_DISABLE_XONXOFF
	ttyb.c_iflag &= ~(IXON | IXOFF);
#endif
	ttyb.c_oflag |= OPOST;
	if(linemode == LM_CHAR)
	    ttyb.c_oflag &= ~OPOST;
	ttyb.c_lflag &= ~(ECHO | ISIG | ICANON);
	if(k)
	    ttyb.c_lflag |= ISIG;
#else /* USE_SGTTY */
	ttyb.sg.sg_flags &= ~(ECHO | RAW);
	ttyb.sg.sg_flags |= CBREAK;
	if(linemode == LM_CHAR)
	    ttyb.sg.sg_flags |= RAW;
	if(!k) {
	    static const struct tchars tc_off = { -1, -1, -1, -1, -1, -1 };
	    static const struct ltchars ltc_off = { -1, -1, -1, -1, -1, -1 };
	    ttyb.tc = tc_off;		/* struct assigns */
	    ttyb.ltc = ltc_off;
	}
#endif
	if(!term_set_attr(0, &ttyb)) /* hm */;
    }
}

/*
 * see whether the window size has changed. return 1 if it has.
 * check_winsize() is always called with sockfd == main_sockfd.
 */
int check_winsize()
{
    /*
     * NOTE: When a window gets smaller, we may need to zero line0
     * (to keep it in window). Subsequently expanding the window
     * can then reveal old text we did not clean up, in case our
     * terminal has memory_above.
     */
#ifdef TIOCGWINSZ
    struct winsize z;
    /* if ioctl fails or gives silly values, don't change anything */
    if(ioctl(0, TIOCGWINSZ, &z) == 0 && z.ws_row && z.ws_col > 1 &&
	    (z.ws_row != lines || z.ws_col != cols)) {
	/* NOTE: we assume the bottom line remains on the bottom. */
	line0 += z.ws_row - lines;
	if(line0 < 0) line0 = 0;
	lines = z.ws_row;
	lines_1 = lines - 1;
	cols = z.ws_col;
	cols_1 = cols - !wrapglitch;
	if(col0 >= cols_1) col0 = 0; /* kill the prompt if it's too long */
	send_term_size();
	return 1;
    }
#endif
    return 0;
}
